"""Functions for scraping web data"""

from .webanalysis import correct_url, get_domain, open_url
from .search import search_web
from ..utils.cleaners import join_list_by_colon, split_str_by_colon
from ..utils.basics import results_cols
from ..importers.pdf import read_pdf_url

from typing import List
import json
import requests

import numpy as np
import pandas as pd

import cloudscraper # type: ignore

from trafilatura import fetch_url, extract, extract_metadata # type: ignore

import urllib
from urllib.request import Request, urlopen

from bs4 import BeautifulSoup

from courlan import normalize_url, clean_url, scrub_url # type: ignore

from selenium import webdriver
from selenium.webdriver.common.by import By


headers = {'User-Agent': 'Mozilla/5.0 (X11; Windows; Windows x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/103.0.5060.114 Safari/537.36'}

can_scrape = [
              'frontiersin.org',
              'arxiv.org',
              'springer',
              'nature.com',
              'ieee.org',
              'pubmed.ncbi.',
              'gov/pmc',
              'ssrn.com',
              'heinonline.org',
              'mdpi.com',
              'acm.org',
              'muse.jhu.edu',
              'proquest.com',
              'jstor.org',
              'scholar.google.com'
              ]

def get_final_url(url):

    global headers
    req = Request(url=url, headers=headers)

    resp = urlopen(req)
    return resp.geturl()

def bs_find(tag, content, soup):
    return ('soup.find(attrs={"'+tag+'":"'+content + '"})', soup.find(attrs={tag:content}))

def bs_find_all(tag, content, soup):
    return ('soup.find_all(attrs={"'+tag+'":"'+content + '"})', soup.find_all(attrs={tag:content}))

def bs_name_content(content_tag, soup):
    return ('soup.find(attrs={"name":"'+content_tag + '"}).attrs["content"]', soup.find(attrs={'name':content_tag}).attrs['content'])

def get_url_source(url = 'request_input'):
    
    """Returns the HTTP response for the provided URL. 

    Parameters
    ---------- 
    url : str
        URL of the page to scrape.

    Returns
    -------
    response : object
        HTTP response object from cloudscraper.
    """
    
    # Requesting URL from user input if none given
    if url == 'request_input':
        url = input('URL: ')
    
    # Correcting errors in URL (e.g. missing HTTPS prefix)
    url = correct_url(url)
   
    # Trying to scrape site using CloudScraper to avoid Cloudflare protection
    try:
        scraper = cloudscraper.create_scraper()
        response = scraper.get(url)
        return response
    
    # Handling errors
    except:
        raise ValueError('Scraper failed')

def url_to_soup(url = 'request_input'):
    
    if url == 'request_input':
        url = input('URL: ')
    
    response = get_url_source(url)
    
    return BeautifulSoup(response.text)

def check_bad_url(url: str = 'request_input') -> bool:
    
    """
    Checks if URL leads to a real website.
    
    Parameters
    ----------
    url : str
        URL to check. Defaults to requesting from user input.
    
    Returns
    -------
    result : bool
        whether the URL is a bad link (True == bad).
    """
    
    # Requesting URL from user input if none given
    if url == 'request_input':
        url = input('URL: ')
    
    # Correcting URL if needed (e.g., adding missing HTTPS prefix)
    url = correct_url(url)
    
    # Trying to make request to URL. If this is successful, URL is not bad; returns False.
    try:
        requests.get(url)
        return False
    
    except:
        
        return True
 
def correct_link_errors(url: str, source_domain: str = None) -> str: # type: ignore
    
    """Checks for errors in a link and corrects them. Returns a corrected link as a string.
    
    Parameters
    ---------- 
    url : str 
        URL to correct.
    source_domain : str 
        the domain or URL where the link was found.

    Returns
    -------
    url : str 
        a corrected URL."""
    
    # Source domain defaults to None. Changing this to an empty string for type handling.
    if source_domain == None:
        source_domain = ''
    
    # Ensuring source_domain is a domain, not a full URL
    source_domain = get_domain(source_domain)
    
    # Checking if domain is valid and retrieving domain if so
    domain_check = get_domain(url)
    
    # If domain check fails
    if (domain_check == '') or (domain_check == None) or ('.' not in domain_check):
        
        # Checking if URL is bad. If true, cleaning URL
        if check_bad_url(url) == True:

            url = url.strip().strip('/').strip()
            url = clean_url(url)
            url = scrub_url(url)
            
            # Checking if URL is still bad. If true, appending source domain to start of URL
            if check_bad_url(url) == True:
                url = source_domain + '/' + url
            
            # Correcting URL if needed (e.g., adding missing HTTPS prefix)
            url = correct_url(url)
    
    return url

def scrape_url_html(url = 'request_input') -> str:
    
    """Scrapes raw HTML code from the provided URL. 

    Parameters
    ---------- 
    url : str
        URL of the page to scrape.

    Returns
    -------
    HTML : str
        the returned HTML as a string."""
    
    # Requesting URL from user input if none given
    if url == 'request_input':
        url = input('URL: ')
    
    # Correcting errors in URL (e.g. missing HTTPS prefix)
    url = correct_url(url = url)
    
    # Retrieving HTTP response
    response = get_url_source(url = url)
    
    # Extracting HTML
    result = response.text
    
    return result

def scrape_url_metadata(url = 'request_input') -> dict:
    
    """Scrapes metadata from the provided URL. 

    Parameters
    ---------- 
    url : str 
        URL of the page to scrape.

    Returns
    -------
    metadata : dict
        the URL's metadata as a dictionary."""
    
    # Requesting URL from user input if none given
    if url == 'request_input':
        url = input('URL: ')
    
    # Correcting errors in URL (e.g. missing HTTPS prefix)
    url = correct_url(url = url)
    
    # Using trafilatura to fetch site data
    downloaded = fetch_url(url = url)
    
    # Creating empty result variable to avoid errors
    result = None
    
    # Extracting metadata and assigning to result variable
    if downloaded != None:
        result = extract_metadata(downloaded)
        
        if result != None:
            
            # Converting result to dictionary
            result = result.as_dict()
            
        else:
            result = {}
    else:
        result = {}
        
    return result

def scrape_url_rawtext(url = 'request_input') -> str:
    
    """Scrapes raw text from the provided URL. 

    Parameters
    ---------- 
    url : str 
        URL of the page to scrape.

    Returns
    -------
    rawtext : str
        the URL's raw text as a string."""
    
    # Requesting URL from user input if none given
    if url == 'request_input':
        url = input('URL: ')
    
    # Correcting errors in URL (e.g. missing HTTPS prefix)
    url = correct_url(url = url)
    
    # Using trafilatura to fetch site data
    downloaded = fetch_url(url = url)
    
    if downloaded != None:
        
        # Extracting raw text and assigning to result variable
        result = extract(downloaded, include_tables=True, include_comments=True, include_images=True, include_links=True)
    
    return result

def scrape_url_xml(url = 'request_input') -> str:
    
    """Scrapes data from the provided URL and returns in XML format. 

    Parameters
    ---------- 
    url : str 
        URL of the page to scrape.

    Returns
    -------
    xml : str
        the URL's data as an XML-formatted string."""
    
    # Requesting URL from user input if none given
    if url == 'request_input':
        url = input('URL: ')
    
    # Correcting errors in URL (e.g. missing HTTPS prefix)
    url = correct_url(url = url)
    
    # Using trafilatura to fetch site data
    downloaded = fetch_url(url)
    if downloaded != None:
        
        # Extracting data and assigning to result variable
        result = extract(downloaded, output_format="xml", include_tables=True, include_comments=True, include_images=True, include_links=True)
    else:
        result = ''

    return result

def scrape_url_json(url = 'request_input') -> str:
    
    """Scrapes data from the provided URL and returns in JSON format. 

    Parameters
    ---------- 
    url : str 
        URL of the page to scrape.

    Returns
    -------
    json : str
        the URL's data as a JSON-formatted string."""
    
    # Requesting URL from user input if none given
    if url == 'request_input':
        url = input('URL: ')
    
    # Correcting errors in URL (e.g. missing HTTPS prefix)
    url = correct_url(url = url)
    
    # Using trafilatura to fetch site data
    downloaded = fetch_url(url)
    if downloaded != None:
        
        # Extracting data and assigning to result variable
        result = extract(downloaded, output_format="json", include_tables=True, include_comments=True, include_images=True, include_links=True)
    else:
        result = ''
        
    return result

def scrape_url_csv(url = 'request_input'):
    
    """Scrapes data from the provided URL and returns in CSV format. 

    Parameters
    ---------- 
    url : str 
        URL of the page to scrape.

    Returns
    -------
    csv : str
        the URL's data as a CSV-formatted string."""
    
    # Requesting URL from user input if none given
    if url == 'request_input':
        url = input('URL: ')
    
    # Correcting errors in URL (e.g. missing HTTPS prefix)
    url = correct_url(url = url)
    
    # Using trafilatura to fetch site data
    downloaded = fetch_url(url)
    if downloaded != None:
        
        # Extracting data and assigning to result variable
        result = extract(downloaded, output_format="csv", include_tables=True, include_comments=True, include_images=True, include_links=True)
    
    return result

def scrape_url_to_dict(url = 'request_input') -> dict:
    
    """Scrapes data from provided URL and returns as a dictionary. 

    Parameters
    ---------- 
    url : str 
        URL of the page to scrape.

    Returns
    -------
    result : dict 
        the URL's data as a dictionary."""
    
    # Retrieving site data as a JSON-formatted string
    result_json = scrape_url_json(url = url)
    
    # Avoiding errors if no result retrieved
    if (result_json == None) or ((type(result_json) is str) and len(result_json) == 0):
        result = {}
    
    else:
        # Converting JSON to dictionary
        try:
            result = json.loads(result_json)
        except:
            result = {}
    
    return result
    
def scrape_url_links(url = 'request_input') -> List[str]:
    
    """Scrapes links from provided URL and returns as a list. 

    Parameters
    ---------- 
    url : str 
        URL of the page to scrape.

    Returns
    -------
    links : list 
        the URL's links as a list."""
    
    # Requesting URL from user input if none given
    if url == 'request_input':
        url = input('URL: ')
    
    # Scraping data
    scraped_data = scrape_url_html(url = url)
    
    # Making HTML soup
    soup = BeautifulSoup(scraped_data, "html.parser")
    
    # Selecting dividers
    href_select = soup.select("a")  
    
    # Extracting links as list and applying corrections 
    links = [correct_link_errors(source_domain = url, url = i['href']) for i in href_select if 'href' in i.attrs] # type: ignore
    
    return links
 
def scrape_url(url = 'request_input', parse_pdf = True, output: str = 'dict'):
    
    """Scrapes data from URL. Returns any HTML code, text, links, and metatdata found. 
    
    Defaults to returning a dictionary.
    
    Parameters
    ---------- 
    url : str 
        URL of the page to scrape.
    parse_pdf : bool
        whether to detect PDFs and parse them using PDF parser.

    Returns
    -------
    result : object
        the result in a user-selected format.
    """
    
    # Requesting URL from user input if none given
    if url == 'request_input':
        url = input('URL: ')
        
    # Cleaning URL
    url = url.strip("'").strip('"').strip()
    
    # Running scrapers depending on the format specified
    
    output = output.lower().strip()
    
    if output == 'html':
        result = scrape_url_html(url = url)
    
    if (output == 'raw') or (output == 'rawtext') or (output == 'raw text'):
        result = scrape_url_rawtext(url = url)
    
    if output == 'xml':
        result = scrape_url_xml(url = url)
    
    if output == 'json':
        result = scrape_url_json(url = url)
    
    if output == 'csv':
        result = scrape_url_csv(url = url)
    
    # If the output format selected is dictionary, scrapes all data available
    if output == 'dict':
        

        result = {}


        # Running main scraper
        try:
            result = scrape_url_to_dict(url = url)
        except:
            pass
        
        # Appending html to output dict
        try:
            result['html'] = scrape_url_html(url = url)
        except:
            result['html'] = ''
        
        current_url = url

        if ('html' in result.keys()) and (result['html'] is not None):
                soup = BeautifulSoup(result['html'], "html.parser")
                try:
                    href_select = soup.select("a")
                    links = [correct_link_errors(source_domain = current_url, url = i['href']) for i in href_select if 'href' in i.attrs] # type: ignore
                except:
                    links = []
                result['links'] = links

                try:
                    html_find = soup.find('html')
                    html_attrs = html_find.attrs # type: ignore
                    if 'lang' in html_attrs:
                        lang = html_attrs['lang']
                    else:
                        lang = None
                except:
                    lang = None
                if ('language' not in result.keys()) or (result['language'] is None) or (result['language'] == ''):
                    result['language'] = lang

                try:
                    head_find = soup.find('head')
                    title_find = head_find.find('title') # type: ignore
                    if title_find is not None:
                        if 'contents' in title_find.__dict__.keys():
                            title = str(title_find.contents).replace('[','').replace(']','').replace('"','').replace("'","").strip() # type: ignore
                        else:
                            title = None
                except:
                    title = None
                    
                if ('title' not in result.keys()) or (result['title'] is None) or (result['title'] == ''):
                    result['title'] = title
        
        # If parse_pdf is selected, check if URL is PDF and parse
        
        if parse_pdf == True:
            
            if url.endswith('.pdf') == True:
                
                # Running PDF downloader and parser
                pdf_parsed = read_pdf_url(url = url)
                
                # Appending result
                if ('title' not in result.keys()) or (result['title'] is None) or (result['title'] == '') or (result['title'] == 'Just a moment...'):
                    result['title'] = pdf_parsed['title']
                
                if ('author' not in result.keys()) or (result['author'] is None) or (result['author'] == '') or (result['author'] == []) or (result['author'] == '[]'):
                    result['author'] = pdf_parsed['authors']

                result['raw_text'] = pdf_parsed['raw']
                result['text'] = pdf_parsed['full_text']

                if ('date' not in result.keys()) or (result['date'] is None) or (result['date'] == ''):
                    result['date'] = pdf_parsed['date']
                
                result['links'] = pdf_parsed['links']
                result['format'] = 'PDF'
                result['type'] = 'document'
        
        # Scraping URL metadata using trafilatura
        metadata = scrape_url_metadata(url = url)
        if metadata != None:
            for key in metadata.keys():
                if key not in result.keys():
                    result[key] = metadata[key]
        
        # Appending URL
        result['url'] = url
        
    return result

def scrape_urls_list(urls: list, parse_pdf = True, output = 'dataframe'):
    
    """Scrapes list of URLs. Returns any HTML code, text, links, and metatdata found. 
    
    Defaults to returning a dataframe.
    
    Parameters
    ---------- 
    url : str 
        URL of the page to scrape.
    parse_pdf : bool
        whether to detect PDFs and parse them using PDF parser.

    Returns
    -------
    result : object 
        the result in a user-selected format."""
    
    # Initialising dictionary for results
    output_dict = {}
    
    # Iterating through links and scraping
    for link in urls:
        output_dict[link] = scrape_url(url = link, parse_pdf = parse_pdf, output = 'dict')
    
    # If selected output is dataframe, converting to dataframe
    if output == 'dataframe':
        output = pd.DataFrame.from_dict(output_dict).T # type: ignore
    
    else:
        output = output_dict
    
    return output

def scrape_dynamic_page(url = 'request_input') -> dict:
    
    """Scrapes dynamic webpages using provided URL. Returns a dictionary of data. Uses Selenium.
    
    Parameters
    ---------- 
    url : str 
        URL of the page to scrape.

    Returns
    -------
    result : dict 
        the result as a dictionary."""
    
    # Initialising dictionary for results
    if url == 'request_input':
        url = input('URL: ')
    
    # Correcting errors in URL (e.g. missing HTTPS prefix)
    url = correct_url(url)
    
    # Initialising Selenium webdriver
    driver = webdriver.Chrome()
    
    # Retrieving data
    driver.get(url)
    
    # Selecting CSS elements
    items = driver.find_elements(By.CSS_SELECTOR, ".grid a[data-testid='link']")
    
    # Iterating through results, retrieving attributes, and adding to output dictionary
    output_dict = {}
    index = 0
    
    for item in items:
        
        name = item.accessible_name
        if name == None:
            try:
                name = item.get_attribute('name')
            except:
                name = None
        
        if name == None:
            name = index
            index += 1
        
        # Retrieving text
        text = item.text
        output_dict[name] = text
    
    return output_dict

def scrape_google_search(query: str = 'request_input'):

    """Fetches and parses Google Search results. Offers an alternative to using Google's API. 
    
    Requires user to manually copy and paste Google search source code to avoid Google's CAPCHA system.
    
    Parameters
    ----------
    query : str
        query to search. Defaults to requesting from user input.
    """
    
    print('This function will open your web browser. When prompted, copy and paste the code that appears into the space provided\n')
    
    # Launching web search
    search_web(
                query = query,
                search_engine = 'Google',
                view_source = True
                )
    
    # Requesting search source code from user
    html = input('Search page code: ')
    
    # Parsing result
    output = parse_google_result(html)  # type: ignore
    
    return output

def crawler_scraper(current_url: str, full: bool) -> tuple:
    
    """Scraper used by web crawler. Returns result as a tuple. 
    
    Parameters
    ---------- 
    current_url : str 
        URL of the page to scrape.
    full : bool 
        whether to run complete scrape.

    Returns
    -------
    result : tuple 
        the result as a tuple containing BeautifulSoup object, the scraped data, and links."""
    
    # If full is selected, runs a complete scrape using tools
    if full == True:
        
        # Tries to scrape URL using standard scraper function
        try:
            scraped_data = scrape_url(current_url)
        
        # If scraper fails, runs cleaners on URL and tries to run it again
        except:
            current_url = clean_url(current_url)
            current_url = scrub_url(current_url)
            current_url = normalize_url(current_url)
            current_url = correct_url(current_url)

            try:
                scraped_data = scrape_url(current_url)

            except:
                
                # If standard scraper fails twice, tries to use cloudscraper's basic scraper
                try:
                    scraper = cloudscraper.create_scraper()
                    res = scraper.get(current_url)
                    scraped_data['html'] = res.content # type: ignore
                    
                except:
                    scraped_data['html'] = '' # type: ignore
    
    # If full is not selected, runs cloudscraper's basic scraper
    else:
        
        try:
            
            scraper = cloudscraper.create_scraper()
            res = scraper.get(current_url)
            scraped_data['html'] = res.content # type: ignore
            
        except:
            scraped_data['html'] = '' # type: ignore
        
    # Making HTML soup
    soup = BeautifulSoup(scraped_data['html'], "html.parser") # type: ignore
    
    # Selecting dividers
    href_select = soup.select("a")  
    
    # Extracting links
    links = [correct_link_errors(source_domain = current_url, url = i['href']) for i in href_select if 'href' in i.attrs]       # type: ignore    
    
    # Returning results as tuple
    return (soup, scraped_data, links)

def scrape_frontiers(url):

    if ('frontiersin.org' not in url) and ('doi.org' not in url):
        raise ValueError('URL must be for a Frontiers webpage')
    
    res = scrape_url(url = url, parse_pdf = False, output = 'html')
    try:
        soup = BeautifulSoup(res,'lxml') # type: ignore
    except:
        pass

    try:
        meta = soup.find_all('meta')
    except:
        meta = ''

    try:
        doi = soup.find(attrs = {'name':"citation_doi"}).attrs['content'] # type: ignore
    except:
        doi = None
    
    try:
        keywords = soup.find(attrs = {'name':"citation_keywords"}).attrs['content'].split('; ') # type: ignore
    except:
        keywords = None
    
    item_type = 'article'
    
    try:
        link = soup.find(attrs = {'property':"og:url"}).attrs['content'] # type: ignore
    except:
        link = None
    
    try:
        date = soup.find(attrs={'name':"citation_publication_date"}).attrs['content'] # type: ignore
    except:
        date = None

    try:
        source = soup.find(attrs = {'name':"citation_journal_title"}).attrs['content'] # type: ignore
    except:
        source = None

    try:
        publisher = soup.find(attrs = {'name':"citation_publisher"}).attrs['content'] # type: ignore
    except:
        publisher = None

    repository = 'Frontiers'

    try:
        title = soup.find(attrs = {'name':"citation_title"}).attrs['content'] # type: ignore
    except:
        title = None

    try:
        description = soup.find(attrs = {'name':"description"}).attrs['content'] # type: ignore
    except:
        description = None

    try:
        abstract = soup.find(attrs={'name':"citation_abstract"}).attrs['content'] # type: ignore
    except:
        abstract = None

    authors = []
    try:
        alist = soup.find_all(attrs={'name':"citation_author"})
        for name in alist:
            authors.append(name.attrs['content'])
    except:
        pass
        
    
    paras = []
    refs = []

    try:
        soup.find_all('p')

        for i in soup.find_all('p'):

            if ('<p class="References' in str(i)):
                refs.append(i.text.strip())

            if (
                ('<span' not in str(i)) 
                and ('<p class="References' not in str(i))
                ):
                paras.append(i.text.replace('\n', '').replace('\r', '').replace('  ', '').replace('   ', '').strip())
    except:
        pass

    full_text = ' \n\n '.join(paras)
    
    global results_cols
    result = pd.DataFrame(columns = results_cols, dtype=object)

    result.loc[0, 'title'] = title
    result.loc[0, 'authors'] = authors
    result.loc[0, 'date'] = date
    result.loc[0, 'source'] = source
    result.loc[0, 'publisher'] = publisher
    result.loc[0, 'type'] = item_type
    result.loc[0, 'keywords'] = keywords
    result.loc[0, 'abstract'] = abstract
    result.loc[0, 'repository'] = repository
    result.loc[0, 'doi'] = doi
    result.loc[0, 'link'] = link
    result.loc[0, 'full_text'] = full_text
    result.loc[0, 'citations'] = refs
    
    return result

def scrape_arxiv(url):
    
    if ('arxiv.org' not in url) and ('doi.org' not in url):
        raise ValueError('URL must be for a Arxiv webpage')
    
    try:
        res = scrape_url(url = url, parse_pdf = False, output = 'html')
        soup = BeautifulSoup(res,'lxml', features="xml") # type: ignore
    except:
        soup = BeautifulSoup()

    try:
        meta = soup.find_all('meta')
    except:
        meta = []

    try:
        doi = soup.find(attrs = {'class':"tablecell arxivdoi"}).a.attrs['href'] # type: ignore
    except:
        doi = None

    try:
        keywords = soup.find(attrs = {'class':"tablecell subjects"}).text.replace('\n', '').split('; ') # type: ignore
    except:
        keywords = None
        
    item_type = 'article'
    
    try:
        link = soup.find(attrs = {'property':"og:url"}).attrs['content'] # type: ignore
    except:
        link = None
        
    try:
        date = soup.find(attrs = {'class': "dateline"}).text.replace('\n', '').replace('  ', '').replace('[', '').replace(']', '').replace('Submitted on ', '') # type: ignore
    except:
        date = None
        
    source = 'arxiv'
    publisher = 'arxiv'
    repository = 'arxiv'
    title = []
    # descriptions = []
    abstract = []
    authors = []


    for i in meta:

        string = str(i)
        
        if 'property="og:site_name"/>' in string:
            string = string.replace('<meta content="', '').replace('" ', '').replace('property="og:site_name"/>', '')
            domain = string

        if 'property="og:title"/>' in string:
            string = string.replace('<meta content="', '').replace('property="og:title"/>', '').replace('" ', '')
            title.append(string)

        if 'property="og:description"/>' in string:
            string = string.replace('<meta content="', '').replace('property="og:description"/>', '').replace('" ', '')
            string = string
            abstract.append(string)


    try:
        title = title[0]
    except:
        pass
    
    try:
        abstract = abstract[0]
    except:
        pass

    authors = []
    try:
        string_list = soup.find(attrs={'class':'authors'}).find_all('a') # type: ignore
    
        for item in string_list:
            string = str(item).replace('<', ' <').replace('>', '> ').replace('</a> ', '').replace(' <a href="https://arxiv.org/search/cs?searchtype=author&amp;query=', '').strip()
            string = string.split('> ')[1]
            authors.append(string)
    except:
        pass

    global results_cols
    result = pd.DataFrame(columns = results_cols, dtype=object)

    result.at[0, 'title'] = title
    result.at[0, 'authors'] = authors
    result.at[0, 'date'] = date
    result.at[0, 'source'] = source
    result.at[0, 'publisher'] = publisher
    result.at[0, 'type'] = item_type
    result.at[0, 'keywords'] = keywords
    result.at[0, 'abstract'] = abstract
    result.at[0, 'repository'] = repository
    result.at[0, 'doi'] = doi
    result.at[0, 'link'] = link
    
    return result

def scrape_springer(url):
    
    if ('springer' not in url) and ('doi.org' not in url):
        raise ValueError('URL must be for a Springer webpage')
    
    try:
        res = scrape_url(url = url, parse_pdf = False, output = 'html')
        soup = BeautifulSoup(res,'lxml')  # type: ignore
    except:
        soup = BeautifulSoup()

    try:
        meta = soup.find_all('meta')
    except:
        meta = []

    doi = []
    keywords = []
    item_type = item_type= []
    link = []
    date = []
    source = []
    publisher = []
    titles = []
    descriptions = []
    authors = []


    for i in meta:

        string = str(i)

        if 'name="citation_doi"/>' in string:
            string = string.replace('<meta content="', '').replace('name="citation_doi"/>', '').replace('" ', '')
            doi.append(string)

        if 'keyword' in string:
            string = string.replace('<meta content="', '').replace('property="og:type"/>', '').replace('" ', '')
            keywords.append(string)

        if 'property="og:type"/>' in string:
            string = string.replace('<meta content="', '').replace('property="og:type"/>', '').replace('" ', '').lower()
            item_type.append(string)

        if 'name="citation_fulltext_html_url"/>' in string:
            string = string.replace('<meta content="', '').replace('name="citation_fulltext_html_url"/>', '').replace('" ', '')
            link.append(string)

        if 'name="citation_publication_date"/>' in string:
            string = string.replace('<meta content="', '').replace('name="citation_publication_date"/>', '').replace('" ', '')
            date.append(string)

        if 'name="citation_inbook_title"' in string:
            string = string.replace('<meta content="', '').replace('name="citation_inbook_title"/>', '').replace('" ', '')
            source.append(string)

        if 'name="citation_publisher"/>' in string:
            string = string.replace('<meta content="', '').replace('name="citation_publisher"/>', '').replace('" ', '')
            publisher.append(string)

        if 'name="citation_title"/>' in string:
            string = string.replace('<meta content="', '').replace('name="citation_title"/>', '').replace('" ', '')
            titles.append(string)

        if 'name="description"/>' in string:
            string = string.replace('<meta content="', '').replace('name="description"/>', '').replace('" ', '')
            string = string
            descriptions.append(string)

        if '"citation_author"' in string:
            string = string.replace('<meta content="', '').replace('name="citation_author"/>', '').replace('" ', '')
            string = string
            authors.append(string)

    try:
        doi = doi[0]
    except:
        pass

    try:
        keywords = keywords[0]
    except:
        pass

    try:
        item_type = item_type[0]
    except:
        pass

    try:
        link = link[0]
    except:
        pass

    try:
        date = date[0]
    except:
        pass

    try:
        source = source[0]
    except:
        pass

    try:
        publisher = publisher[0]
    except:
        pass


    try:
        title = titles[0]
    except:
        pass
    
    try:
        notes = soup.find_all('section')[1].find_all('p')
    except:
        notes = []

    refs = []
    citations_data = []

    for i in notes:

        string = str(i)
        string = string.replace('<p>', '').replace('</p>', '').replace('[', '').replace(']', '')
        refs.append(string)

        try:
            citations_data.append(i.select('a')[0].attrs['href'])

        except:
            pass

    try:
        sections = soup.find_all('section')
        abstract = str(sections[0].find_all('p')).replace('<p>', '').replace('</p>', '').replace('[', '').replace(']', '')
    except:
        sections = []
        abstract = None

    global results_cols
    result = pd.DataFrame(columns = results_cols, dtype=object)
    
    result.at[0, 'title'] = title
    result.at[0, 'authors'] = authors
    result.at[0, 'date'] = date
    result.at[0, 'source'] = source
    result.at[0, 'publisher'] = publisher
    result.at[0, 'type'] = item_type
    result.at[0, 'keywords'] = keywords
    result.at[0, 'abstract'] = abstract
    result.at[0, 'citations'] = refs
    result.at[0, 'citations_data'] = citations_data
    result.at[0, 'repository'] = 'Springer'
    result.at[0, 'doi'] = doi
    result.at[0, 'link'] = link
    
    return result

def scrape_nature(url = 'request_input'):
    
    if url == 'request_input':
        url = input('URL: ')
    
    if type(url) !=str:
        raise TypeError('URL must be a string')
    
    if ('nature.com' not in url) and ('doi.org' not in url):
        raise ValueError('URL must be for a Nature webpage')
    
    try:
        res = scrape_url(url = url, parse_pdf = False, output = 'html')
        soup = BeautifulSoup(res, 'lxml') # type: ignore
    except:
        soup = BeautifulSoup()

    try:
        all_scripts = soup.find_all('script')
        for i in all_scripts:
            if 'type="application/ld+json">{"mainEntity"' in str(i):
                script_json = str(i).replace('<script type="application/ld+json">', '').replace('</script>', '')
    except:
        script_json = '{}'

    try:
        main = json.loads(script_json)['mainEntity']

        title = main['headline']
        abstract = main['description']
        date = main['datePublished']
        doi = main['sameAs']
        keywords = main['keywords']
        source = main['isPartOf']['name']
        publisher = main['publisher']['name']
        item_type = 'article'

        authors = []
        for i in main['author']:
            authors.append(i['name'])
    except:
        title = None
        abstract = None
        date = None
        doi = None
        keywords = None
        source = None
        publisher = None
        item_type = 'article'
        authors = None

    try:
        for i in soup.find_all('link'):
            string = str(i)
            if 'canonical' in string:
                link = string.replace('<link href="', '').replace('rel="canonical"/>', '').replace('" ', '')
    except:
        link = None

    try:
        meta = soup.find_all('meta')

        citations = []
        citations_links = []
        for i in meta:
            if 'name="citation_reference"/>' in str(i):
                citation = i['content'].split('; ')
                cit_dict = {}
                for i in range(0, len(citation)):
                    if '=' in citation[i]:
                        pair = citation[i].split('=')
                        cit_dict[pair[0]] = pair[1]
                        if pair[0] == 'citation_doi':
                            citations_links.append(pair[1])
                    else:
                        cit_dict['item_title'] = citation[i].replace('\n', '').replace('  ', '').replace('   ', '').strip()
                
                citations.append(cit_dict)
    except:
        citations = None
    
    global results_cols
    result = pd.DataFrame(columns = results_cols, dtype=object)
         
    result.loc[0, 'title'] = title
    result.loc[0, 'authors'] = authors
    result.loc[0, 'date'] = date
    result.loc[0, 'source'] = source
    result.loc[0, 'publisher'] = publisher
    result.loc[0, 'type'] = item_type
    result.loc[0, 'keywords'] = keywords
    result.loc[0, 'abstract'] = abstract
    result.loc[0, 'citations'] = citations
    result.loc[0, 'citations_data'] = citations_links
    result.loc[0, 'repository'] = 'Nature'
    result.loc[0, 'doi'] = doi
    result.loc[0, 'link'] = link
    
    return result

def scrape_ieee(url):

    if type(url) !=str:
        raise TypeError('URL must be a string')
    
    if ('ieee.org' not in url) and ('doi.org' not in url):
        raise ValueError('URL must be for an IEEE webpage')
    
    try:
        res = scrape_url(url = url, parse_pdf = False, output = 'html')
        soup = BeautifulSoup(res, 'lxml') # type: ignore
    except:
        soup = BeautifulSoup()

    try:
        meta = soup.find_all('meta')
    except:
        meta = []

    item_type = 'article'
    link = []
    date = []
    source = []
    publisher = 'IEEE'
    titles = []
    descriptions = []
    abstract = []
    authors = []


    for i in meta:

        string = str(i)

        if 'name="parsely-link"' in string:
            string = string.replace('<meta content="', '').replace('name="parsely-link"/>', '').replace('" ', '')
            link.append(string)

        if 'property="parsely-date"/>' in string:
            string = string.replace('<meta content="', '').replace('name="citation_publication_date"/>', '').replace('" ', '')
            date.append(string)

        if 'name="parsely-section"/>' in string:
            string = string.replace('<meta content="', '').replace('name="parsely-section"/>', '').replace('" ', '')
            source.append(string)

        if 'property="og:title"/>' in string:
            string = string.replace('<meta content="', '').replace('property="og:title"/>', '').replace('" ', '')
            titles.append(string)

        if 'id="meta-description"' in string:
            string = string.replace('<meta content="', '').replace('id="meta-description" name="Description"/>', '').replace('" ', '')
            string = string
            descriptions.append(string)
        
        if 'property="og:description"/>' in string:
            string = string.replace('<meta content="', '').replace('property="og:description"/>', '').replace('" ', '')
            string = string
            abstract.append(string)

        if 'name="parsely-author"/>' in string:
            string = string.replace('<meta content="', '').replace('name="parsely-author"/>', '').replace('" ', '')
            string = string
            authors.append(string)

    
    try:
        link = link[0]
    except:
        pass

    try:
        date = date[0]
    except:
        pass

    try:
        source = source[0]
    except:
        pass

    try:
        descriptions = descriptions[0]
    except:
        pass     
        
    try:
        abstract = abstract[0]
    except:
        pass

    try:
        title = titles[0]
    except:
        pass

    global results_cols
    result = pd.DataFrame(columns = results_cols, dtype=object)

    result.loc[0, 'title'] = title
    result.loc[0, 'authors'] = authors
    result.loc[0, 'date'] = date
    result.loc[0, 'source'] = source
    result.loc[0, 'publisher'] = publisher
    result.loc[0, 'type'] = item_type
    result.loc[0, 'description'] = descriptions
    result.loc[0, 'abstract'] = abstract
    result.loc[0, 'repository'] = 'IEEE'
    result.loc[0, 'link'] = link
    
    return result

def scrape_pubmed(url):

    if type(url) !=str:
        raise TypeError('URL must be a string')
    
    if ('pubmed.ncbi.' not in url) and ('doi.org' not in url):
        raise ValueError('URL must be for a PubMed webpage')
    
    try:
        res = scrape_url(url = url, parse_pdf = False, output = 'html')
        soup = BeautifulSoup(res, 'lxml') # type: ignore
    except:
        soup = BeautifulSoup()

    try:
        meta = soup.find_all('meta')
    except:
        meta = []

    try:
        doi = soup.find(attrs={'name':'citation_doi'}).attrs['content'] # type: ignore
    except:
        doi = None
    item_type = 'article'
    
    try:
        link = soup.find(attrs={'rel':'canonical'}).attrs['href'] # type: ignore
    except:
        link = None
    try:
        date = soup.find(attrs={'name':'citation_date'}).attrs['content'] # type: ignore
    except:
        date = None
    try:
        source = soup.find(attrs={'name':'citation_journal_title'}).attrs['content'] # type: ignore
    except:
        source = None
    try:
        publisher = soup.find(attrs={'name':'citation_publisher'}).attrs['content'] # type: ignore
    except:
        publisher = None
    repository = 'PubMed'
    try:
        title = soup.find(attrs={'name':'citation_title'}).attrs['content'] # type: ignore
    except:
        title = None
    try:
        description = soup.find(attrs={'name':'description'}).attrs['content'] # type: ignore
    except:
        description = None
    try:
        abstract = soup.find(attrs={'class':"abstract-content selected"}).text.replace('\n', '').replace('  ', '').replace('   ', '') # type: ignore
    except:
        abstract = None
    try:
        authors = soup.find(attrs={'name':'citation_authors'}).attrs['content'].split(';') # type: ignore
    except:
        authors = None
    
    citations = []
    citations_data = []

    try:
        refs = soup.find_all(attrs={'class':'references-and-notes-list'})
    except:
        refs = []

    for citation in refs:
        citations.append(citation.text.replace('\n', '').replace('   ', '').replace('  ', '').replace(' .', '.').strip())
        citations_data.append(citation.li.a.attrs['href'])
    
    keywords = []
    try:
        for i in soup.find_all('p'):
            if 'keywords' in str(i).lower():
                keywords = i.text.replace('\n', '').replace('  ', '').replace('   ', '').replace('Keywords:', '').replace('.', '').split('; ')
    except:
        pass
                
    
    global results_cols
    result = pd.DataFrame(columns = results_cols, dtype=object)

    result.loc[0, 'title'] = title
    result.loc[0, 'authors'] = authors
    result.loc[0, 'date'] = date
    result.loc[0, 'source'] = source
    result.loc[0, 'publisher'] = publisher
    result.loc[0, 'type'] = item_type
    result.at[0, 'keywords'] = keywords
    result.loc[0, 'abstract'] = abstract
    result.loc[0, 'description'] = description
    result.loc[0, 'repository'] = repository
    result.loc[0, 'doi'] = doi
    result.loc[0, 'link'] = link
    result.loc[0, 'citations'] = citations
    result.loc[0, 'citations_data'] = citations_data
    
    return result

def scrape_pmc(url):

    if type(url) !=str:
        raise TypeError('URL must be a string')
    
    if ('gov/pmc' not in url) and ('doi.org' not in url):
        raise ValueError('URL must be for a PubMedCentral webpage')
    
    try:
        res = scrape_url(url = url, parse_pdf = False, output = 'html')
        soup = BeautifulSoup(res, 'lxml') # type: ignore
    except:
        soup = BeautifulSoup()

    
    try:
        doi = soup.find(attrs={'name':'citation_doi'}).attrs['content'] # type: ignore
    except:
        doi = None
    try:
        item_type = soup.find(attrs={'property':'og:type'}).attrs['content'] # type: ignore
    except:
        item_type = None
    try:
        link = soup.find(attrs={'rel':'canonical'}).attrs['href'] # type: ignore
    except:
        link = None
    try:
        date = soup.find(attrs={'name':'citation_publication_date'}).attrs['content'] # type: ignore
    except:
        date = None
    try:
        source = soup.find(attrs={'name':'citation_journal_title'}).attrs['content'] # type: ignore
    except:
        source = None
    try:
        publisher = soup.find(attrs={'name':'DC.Publisher'}).attrs['content'] # type: ignore
    except:
        publisher = None
    repository = 'PubMedCentral'
    try:
        title = soup.find(attrs={'name':'citation_title'}).attrs['content'] # type: ignore
    except:
        title = None
    try:
        description = soup.find(attrs={'property':'og:description'}).attrs['content'] # type: ignore
    except:
        description = None
    try:
        keywords = soup.find(attrs={'class':"kwd-text"}).text # type: ignore
    except:
        keywords = None
    try:
        abstract = soup.find(attrs={'class':"p p-first-last"}).text # type: ignore
    except:
        abstract = None
    
    authors = []
    try:
        for author in soup.find_all(attrs={'name':'citation_author'}):
            authors.append(author.attrs['content'])
    except:
        pass
    
    citations = []
    citations_data = []

    refs = []
    try:
        refs = soup.find(attrs={'class':"ref-list-sec"})
    except:
        pass

    for i in refs: # type: ignore
        citations.append(i.text.replace('\n', '').replace('..', '.').replace('  ', '').replace('   ', '').replace('[PubMed]', '').replace('[Google Scholar]', '').replace('[PMC free article]', '')) # type: ignore

    try:
        for i in refs.find_all(attrs={'target':"_blank"}): # type: ignore
            citations_data.append(i.attrs['href'])
    except:
        pass
    
    full_text = []
    try:
        for i in soup.find_all('p'):
            if 'class="p p-' in str(i):
                full_text.append(i.text.replace('\n', '').replace('  ', '').replace('   ', ''))
    except:
        pass

    full_text = '\n\n'.join(full_text)
    
    global results_cols
    result = pd.DataFrame(columns = results_cols, dtype=object)

    result.loc[0, 'title'] = title
    result.loc[0, 'authors'] = authors
    result.loc[0, 'date'] = date
    result.loc[0, 'source'] = source
    result.loc[0, 'publisher'] = publisher
    result.loc[0, 'type'] = item_type
    result.loc[0, 'keywords'] = keywords
    result.loc[0, 'abstract'] = abstract
    result.loc[0, 'description'] = description
    result.loc[0, 'full_text'] = full_text
    result.loc[0, 'repository'] = repository
    result.loc[0, 'doi'] = doi
    result.loc[0, 'link'] = link
    result.loc[0, 'citations'] = citations
    result.loc[0, 'citations_data'] = citations_data

    return result

def scrape_ssrn(url = 'request_input'):

    if url == 'request_input':
        url = input('URL: ')
    
    if type(url) !=str:
        raise TypeError('URL must be a string')
    
    if ('ssrn.com' not in url) and ('doi.org' not in url):
        raise ValueError('URL must be for an SSRN webpage')
    
    try:
        res = scrape_url(url = url, parse_pdf = False, output = 'html')
        soup = BeautifulSoup(res, 'lxml') # type: ignore
    except:
        soup = BeautifulSoup()

    try:
        doi = soup.find(attrs={'name':'citation_doi'}).text # type: ignore
    except:
        doi = None

    item_type = 'article'

    try:
        link = soup.find(attrs={'rel':'canonical'}).attrs['href'] # type: ignore
    except:
        link = None
    
    try:
        date = soup.find(attrs={'name':'citation_publication_date'}).attrs['content'] # type: ignore
    except:
        date = None
    try:
        source = soup.find(attrs={'class': 'btn-link'}).text # type: ignore
    except:
        source = None
    publisher = 'SSRN'
    repository = 'SSRN'
    try:
        title = soup.find(attrs={'name': 'citation_title'}).attrs['content'] # type: ignore
    except:
        title = None
    try:
        description = soup.find(attrs={'name': 'description'}).attrs['content'] # type: ignore
    except:
        description = None
    try:
        keywords = soup.find(attrs={'name': 'citation_keywords'}).attrs['content'] # type: ignore
    except:
        keywords = None
    try:
        abstract = soup.find(attrs={'class':'abstract-text'}).p.text # type: ignore
    except:
        abstract = None
    
    authors = []
    try:
        for author in soup.find_all(attrs={'name':'citation_author'}):
            authors.append(author.attrs['content'])
    except:
        pass
    
    
    global results_cols
    result = pd.DataFrame(columns = results_cols, dtype=object)

    result.loc[0, 'title'] = title
    result.loc[0, 'authors'] = authors
    result.loc[0, 'date'] = date
    result.loc[0, 'source'] = source
    result.loc[0, 'publisher'] = publisher
    result.loc[0, 'type'] = item_type
    result.loc[0, 'keywords'] = keywords
    result.loc[0, 'abstract'] = abstract
    result.loc[0, 'description'] = description
    result.loc[0, 'repository'] = repository
    result.loc[0, 'doi'] = doi
    result.loc[0, 'link'] = link
    
    return result

def scrape_heinonline(url):

    if type(url) !=str:
        raise TypeError('URL must be a string')
    
    if ('heinonline.org' not in url) and ('doi.org' not in url):
        raise ValueError('URL must be for a HeinOnline webpage')
    
    try:
        res = scrape_url(url = url, parse_pdf = False, output = 'html')
        soup = BeautifulSoup(res, 'lxml') # type: ignore
    except:
        soup = BeautifulSoup()
    
    item_type = 'article'
    repository = 'HeinOnline'

    try:
        title = soup.find(attrs={'class':'cite_title'}).text.strip() # type: ignore
    except:
        title = None
    try:
        description = soup.find(attrs={'name': 'description'}).attrs['content'].replace('\r', '').replace('\n', '').replace('  ', '').replace('   ', '').strip() # type: ignore
    except:
        description = None
    try:
        keywords = soup.find(attrs={'name': 'keywords'}).attrs['content'].replace('\r', '').replace('\n', '').replace('  ', '').replace('   ', '').strip() # type: ignore
    except:
        keywords = None
    try:
        author_res = soup.find(attrs={'class':"Z3988"}).attrs['title'].replace('ctx_ver', ' ').replace('rft', ' ').replace('=', ' ').replace('%', ' ').replace('&', ' ').strip() # type: ignore
    except:
        author_res = ''
    
    auths_to_clean = ''
    for i in author_res.split('_'):
            if '.au' in i:
                auths_to_clean = i

    auths_to_clean = auths_to_clean.split('.')

    authors = []
    for auth in auths_to_clean:
        if 'au ' in auth:
            authors.append(auth.replace('au','').replace('  ','').strip())
    
    try:
        date_res = soup.find(attrs={'class':"Z3988"}).attrs['title'].replace('ctx_ver', ' ').replace('rft', ' ').replace('=', ' ').replace('%', ' ').replace('&', ' ').split('.') # type: ignore
    except:
        date_res = []

    for i in date_res:
        if 'date' in i:
            date = i.replace('date', '').strip()
    
    global results_cols
    result = pd.DataFrame(columns = results_cols, dtype=object)

    result.loc[0, 'title'] = title
    result.loc[0, 'authors'] = authors
    result.loc[0, 'date'] = date
    result.loc[0, 'type'] = item_type
    result.loc[0, 'keywords'] = keywords
    result.loc[0, 'description'] = description
    result.loc[0, 'repository'] = repository
    result.loc[0, 'link'] = url
    
    return result

def scrape_mdpi(url):

    if type(url) !=str:
        raise TypeError('URL must be a string')
    
    if ('mdpi.com' not in url) and ('doi.org' not in url):
        raise ValueError('URL must be for a MDPI webpage')
        
    try:
        res = scrape_url(url = url, parse_pdf = False, output = 'html')
        soup = BeautifulSoup(res, 'lxml') # type: ignore
    except:
        soup = BeautifulSoup()

    try:
        doi = soup.find(attrs={'name':'citation_doi'}).text # type: ignore
    except:
        doi = None
    
    item_type = 'article'
    link = url

    try:
        date = soup.find(attrs={'name':'citation_publication_date'}).attrs['content'] # type: ignore
    except:
        date = None
    try:
        source = soup.find(attrs={'name': 'dc.source'}).attrs['content'] # type: ignore
    except:
        source = None
    try:
        publisher = soup.find(attrs={'name': 'dc.publisher'}).attrs['content'] # type: ignore
    except:
        publisher = None
    
    repository = 'PNAS'

    try:
        title = soup.find(attrs={'name': 'title'}).attrs['content'] # type: ignore
    except:
        title = None
    try:
        description = soup.find(attrs={'name': 'description'}).attrs['content'] # type: ignore
    except:
        description = None
    try:
        abstract = soup.find(attrs={'name': 'description'}).attrs['content'] # type: ignore
    except:
        abstract = None
    
    authors = []
    try:
        for author in soup.find_all(attrs={'name':'dc.creator'}):
            authors.append(author.attrs['content'])
    except:
        pass

    keywords = []
    try:
        for term in soup.find_all(attrs={'name':'dc.subject'}):
            keywords.append(term.attrs['content'])
    except:
        pass


    global results_cols
    result = pd.DataFrame(columns = results_cols, dtype=object)

    result.loc[0, 'title'] = title
    result.loc[0, 'authors'] = authors
    result.loc[0, 'date'] = date
    result.loc[0, 'source'] = source
    result.loc[0, 'publisher'] = publisher
    result.loc[0, 'type'] = item_type
    result.loc[0, 'keywords'] = keywords
    result.loc[0, 'abstract'] = abstract
    result.loc[0, 'description'] = description
    result.loc[0, 'repository'] = repository
    result.loc[0, 'doi'] = doi
    result.loc[0, 'link'] = link

    return result

def scrape_acm(url = 'request_input'):

    if url == 'request_input':
        url = input('URL: ')
    
    if type(url) !=str:
        raise TypeError('URL must be a string')
    
    if ('acm.org' not in url) and ('doi.org' not in url):
        raise ValueError('URL must be for an ACM webpage')
    
    try:
        res = scrape_url(url = url, parse_pdf = False, output = 'html')
        soup = BeautifulSoup(res, 'lxml') # type: ignore
    except:
        soup = BeautifulSoup()
    
    item_type = 'article'
    try:
        link = soup.find(attrs={'property':'og:url'}).attrs['content'] # type: ignore
    except:
        link = None
    try:
        doi = link.replace('https://dl.acm.org/', '') # type: ignore
    except:
        doi = None
    try:
        date = soup.find(attrs={'class':"rlist article-chapter-history-list"}).text.replace('Published: ', '') # type: ignore
    except:
        date = None
    try:
        source = soup.find(attrs={'class':"epub-section__title"}).text # type: ignore
    except:
        source = None
     
    publisher = 'ACM'
    repository = 'ACM'

    try:
        title = soup.find(attrs={'property':'og:title'}).attrs['content'] # type: ignore
    except:
        title = None
    try:
        abstract = soup.find(attrs={'class':"abstractSection abstractInFull"}).p.text.strip() # type: ignore
    except:
        abstract = None
    
    keywords = []
    try:
        chart_terms = soup.find(attrs={'class':"rlist organizational-chart"}).find_all('a') # type: ignore
        
        for i in chart_terms:
            keywords.append(i.text)
    except:
        pass

    authors = []
    try:
        names_txt = soup.find_all(attrs={'class':'author-name'})
        for i in names_txt:
            authors.append(i.attrs['title'])
    except:
        pass
    
    citations = []
    try:
        refs_txt = soup.find_all(attrs={'class':"references__note"})
        citations_data = []
        for i in refs_txt:
            citations.append(i.text.replace('Google Scholar', '').replace('Cross Ref', '').replace('Digital Library', ''))
            citations_data.append(i.find(attrs={'class':'references__suffix'}).a.attrs['href'])
    except:
        pass
    
    global results_cols
    result = pd.DataFrame(columns = results_cols, dtype=object)

    result.loc[0, 'title'] = title
    result.loc[0, 'authors'] = authors
    result.loc[0, 'date'] = date
    result.loc[0, 'source'] = source
    result.loc[0, 'publisher'] = publisher
    result.loc[0, 'type'] = item_type
    result.loc[0, 'keywords'] = keywords
    result.loc[0, 'abstract'] = abstract
    result.loc[0, 'repository'] = repository
    result.loc[0, 'doi'] = doi
    result.loc[0, 'link'] = link
    result.loc[0, 'citations'] = citations
    result.loc[0, 'citations_data'] = citations_data
    
    return result

def parse_muse_from_source(source = 'request_input', link = None):
    
    if source == 'request_input':
        source = input('HTML: ')
    
    if type(source) != str:
        raise TypeError('Source must be a string')

    try:
        soup = BeautifulSoup(source,'lxml')
    except:
        soup = BeautifulSoup()
    
    try:
        authors_html = soup.find_all(attrs={'name':'citation_author'})
        authors = [i.attrs['content'] for i in authors_html]
    except:
        authors = None
        
    try:
        item_type = soup.find(attrs={'class':'type'}).text # type: ignore
    except:
        item_type = None
    
    try:
        link = soup.find(attrs={'name':'citation_fulltext_html_url'}).attrs['content'] # type: ignore
    except:
        link = link
    
    try:
        doi = soup.find(attrs={'name':'citation_doi'}).attrs['content'] # type: ignore
    except:
        doi = None
    
    try:
        date = soup.find(attrs={'name':'citation_year'}).attrs['content'] # type: ignore
    except:
        date = None
    
    try:
        source = soup.find(attrs={'name':"citation_journal_title"}).attrs['content'] # type: ignore
    except:
        source = None
    
    try:
        publisher = soup.find(attrs={'name':'citation_publisher'}).attrs['content'] # type: ignore
    except:
        publisher = None
    
    try:
        title = soup.find(attrs={'name':'citation_title'}).attrs['content'] # type: ignore
    except:
        title = None
        
    try:
        abstract = soup.find(attrs={'class':"abstract"}).text.replace('Abstract:', '').replace('\nAbstract', '').replace('   ', ' ').replace('  ', ' ').strip('\n').strip() # type: ignore
    except:
        abstract = None
        
    repository = 'Project MUSE'
    
    refs_html = soup.find_all(attrs={'name':'citation_reference'})
    try:
        refs_list = [i.attrs['content'] for i in refs_html]
    except:
        refs_list = []
        
    citations = []

    for i in refs_list:
        
        ref_dict = {'authors': []}
        ref_split = i.split('; ')
        
        for entry in ref_split:
            pair = entry.split('=')
            
            if pair[0] == 'citation_author':
                ref_dict['authors'].append(pair[1])
            else:
                ref_dict[pair[0]] = pair[1]

        citation = ''

        if 'authors' in ref_dict.keys():
            citation = citation + ', '.join(ref_dict['authors']) + ' '

        if 'citation_year' in ref_dict.keys():
            citation = citation + '(' + ref_dict['citation_year'] + '),' # type: ignore

        if 'citation_title' in ref_dict.keys():
            citation = citation + ' ' + ref_dict['citation_title'] + '.' # type: ignore

        if 'citation_journal_title' in ref_dict.keys():
            citation = (citation + ' '  # type: ignore
                        + ref_dict['citation_journal_title']
                       )

        if 'citation_volume' in ref_dict.keys():
            citation = (citation + ' '  # type: ignore
                        + ref_dict['citation_journal_title']
                       )

        if 'citation_issue' in ref_dict.keys():
            citation = (citation + '(' + ref_dict['citation_issue']  + ').' ) # type: ignore

        if 'citation_firstpage' in ref_dict.keys():
            citation = (citation + ' pp.' + ref_dict['citation_firstpage']) # type: ignore

        if 'citation_lastpage' in ref_dict.keys():
            citation = (citation + '-' + ref_dict['citation_lastpage'] + '.') # type: ignore

        if 'citation_publisher' in ref_dict.keys():
            citation = (citation + ' ' + ref_dict['citation_publisher'] + '.') # type: ignore

        citation = citation.strip()
        citations.append(citation)
    
    global results_cols
    result = pd.DataFrame(columns = results_cols, dtype=object)

    result.loc[0, 'title'] = title
    result.loc[0, 'authors'] = authors
    result.loc[0, 'date'] = date
    result.loc[0, 'source'] = source
    result.loc[0, 'publisher'] = publisher
    result.loc[0, 'type'] = item_type
    result.loc[0, 'abstract'] = abstract
    result.loc[0, 'repository'] = repository
    result.loc[0, 'doi'] = doi
    result.loc[0, 'link'] = link
    result.loc[0, 'citations'] = citations
    
    return result

def scrape_muse(url = 'request_input'):
    
    if url == 'request_input':
        url = input('URL: ')
    
    if type(url) !=str:
        raise TypeError('URL must be a string')
    
    if ('muse.jhu.edu' not in url) and ('doi.org' not in url):
        raise ValueError('URL must be for a Project MUSE webpage')
    
    try:
        res = scrape_url(url = url, parse_pdf = False, output = 'html')
    except:
        res = ''
    
    result = parse_muse_from_source(source = res, link = url) # type: ignore
    
    return result

def parse_proquest_from_source(source, link = None):

    if type(source) !=str:
        raise TypeError('Source must be a string')

    soup = BeautifulSoup(source,'lxml')
    
    try:
        for i in soup.find_all(attrs={'type':'text/javascript'}):
            if 'pi(' in str(i):
                result = i.text.replace('require(["t5/core/pageinit"], function(pi) ', '').replace(' });', '}').strip('{ ').strip('}')

        result = result.replace('pi(', '').replace('["', '').split('],')
        metadata_txt = [i for i in result if 'DocViewAnalytics' in i]
        meta_json = metadata_txt[0].replace("components/docview/DocViewAnalytics", "").strip(',"')
        metadata = json.loads(meta_json)
    
    except:
        metadata = {
            'primaryAuthor': None, 
            'author2': None, 
            'author3: None,': None, 
            'containsCitedBy': None, 
            'documentLanguage': None, 
            'doi': None, 
            'fulltextAccess': None, 
            'goID': None, 
            'issn': None, 
            'languageOfPublication': None, 
            'pageCount': None, 
            'objectType': None, 
            'isOpenAccess': None, 
            'pubPlace': None, 
            'pubId': None, 
            'pubTitle': None, 
            'year': None, 
            'recordType': None, 
            'searchId': None, 
            'searchQueryTerms': None, 
            'sourceType': None, 
            'subjectIndex1': None, 
            'subjectIndex2': None, 
            'subjectIndex3': None, 
            'subjectIndex4': None, 
            'displayFormatString': None
                }
    
    doi = metadata['doi']
    item_type = metadata['recordType']
    link = link
    date = metadata['year']
    source = metadata['pubTitle']
    repository = 'ProQuest'

    try:
        title = soup.find(attrs={'id':'documentTitle'}).text # type: ignore
    except:
        title = None

    try:
        auths_txt = soup.find(attrs={'class':"truncatedAuthor"}).text.replace('\xa0', '').replace('\\u', ' ').replace('\n', '') # type: ignore
    except:
        auths_txt = ''

    authors = auths_txt.split('.')[0].split(';')
    
    try:
        abstract = soup.find(attrs={'class':"abstractContainer"}).text.replace('\nTranslate', '').replace('Abstract. ', '').replace('Abstract', '').replace('ABSTRACT. ', '').replace('ABSTRACT', '').replace('You have requested "on-the-fly" machine translation of selected content from our databases. This functionality is provided solely for your convenience and is in no way intended to replace human translation. Show full disclaimerNeither ProQuest nor its licensors make any representations or warranties with respect to the translations. The translations are automatically generated "AS IS" and "AS AVAILABLE" and are not retained in our systems. PROQUEST AND ITS LICENSORS SPECIFICALLY DISCLAIM ANY AND ALL EXPRESS OR IMPLIED WARRANTIES, INCLUDING WITHOUT LIMITATION, ANY WARRANTIES FOR AVAILABILITY, ACCURACY, TIMELINESS, COMPLETENESS, NON-INFRINGMENT, MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. Your use of the translations is subject to all use restrictions contained in your Electronic Products License Agreement and by using the translation functionality you agree to forgo any and all claims against ProQuest or its licensors for your use of the translation functionality and any output derived there from. Hide full disclaimer\n\nLonger documents can take a while to translate. Rather than keep you waiting, we have only translated the first few paragraphs. Click the button below if you want to translate the rest of the document.\nTranslate All', '') # type: ignore
    except:
        abstract = None
        
    try:
        full_text_list = soup.find(attrs={'id':"fullTextZone"}).find_all('p') # type: ignore
        full_text_list = [i.text for i in full_text_list]

        full_text = '\n'.join(full_text_list).split('\nReferences\n')
        main_body = full_text[0]
        
        try:
            citations = full_text[1].split('\n')
            citations_data = full_text[1].split(' ')
            citations_data = [i for i in citations_data if (('http' in i) or ('www.') in i)]
        
        except:
            citations = None
            citations_data = None
        
    except:
        main_body = None
        citations = None
        citations_data = None
    
    if (main_body == None) or (abstract == None):
        try:
            main_text = soup.find(attrs={'role':'main'}).find_all('div') # type: ignore
            res_list = [i.find_all('p') for i in main_text if 'ABS' in str(i)]

            try:
                    res_text = [i.text for i in res_list[0]]
                    full_text = '\n'.join(res_text).replace('\nTranslate\n', '')
                    full_text = full_text.split('\nShow less\n')[0].replace('\n\n','').replace('Show less','')
                    split_text = full_text.split('\n\xa0\n')
                    abstract = split_text[0]

                    try:
                        main_body = split_text[1]
                    except:
                        main_body = None

            except:
                    abstract = None
                    main_body = None

        except:
            abstract = None
            main_body = None
    
    if (type(main_body) == str) and (type(abstract) == str):
        main_body = main_body.replace('ABSTRACT', '').replace('Abstract', '').strip().strip('.')
        abstract = abstract.replace('ABSTRACT', '').replace('Abstract', '').strip().strip('.')

    
    global results_cols
    result = pd.DataFrame(columns = results_cols, dtype=object)

    result.loc[0, 'title'] = title
    result.loc[0, 'authors'] = authors
    result.loc[0, 'date'] = date
    result.loc[0, 'source'] = source
    result.loc[0, 'type'] = item_type
    result.loc[0, 'abstract'] = abstract
    result.loc[0, 'full_text'] = main_body
    result.loc[0, 'repository'] = repository
    result.loc[0, 'citations'] = citations
    result.loc[0, 'citations_data'] = citations_data
    result.loc[0, 'doi'] = doi
    result.loc[0, 'link'] = link
    
    return result

def scrape_proquest(url = 'request_input'):
    
    if url == 'request_input':
        url = input('URL: ')
    
    if type(url) !=str:
        raise TypeError('URL must be a string')
    
    if ('proquest.com' not in url) and ('doi.org' not in url):
        raise ValueError('URL must be for a ProQuest webpage')
    
    try:
        res = scrape_url(url = url, parse_pdf = False, output = 'html')
    except:
        res = ''
    
    result = parse_proquest_from_source(source = res, link = url)
    
    return result

def parse_jstor_from_source(source = 'request_input', link = None):

    if source == 'request_input':
        source = input('Source code: ')
    
    if type(source) !=str:
        raise TypeError('Source must be a string')

    if (
        ('unusual traffic activity from your address' in source) 
        or ('unusual traffic activity from your device' in source) 
        or ('detected unusual traffic activity from your network' in source) 
        or ('complete this reCAPTCHA' in source) 
        or ('fill this reCAPTCHA' in source)
        or ('you making the requests and not a robot' in source)
        or ('confirm you are not a robot' in source)
        ):
        raise ValueError('The web browser seems to have been blocked by a bot detection system')
               
    try:
        soup = BeautifulSoup(source,'lxml')
    except:
        soup = BeautifulSoup()
    
    repository = 'JSTOR'
    
    try:
        res = soup.find_all(attrs={'type':'text/javascript'})
        meta_text = [i.text for i in res if 'itemTitle' in str(i)][0]
        meta_list = meta_text.replace('   ', '').replace('  ', '').replace('\n', '').split('{')
        meta_text = [i for i in meta_list if 'itemTitle' in i][0].split('}')[0].strip()
        meta_json = '{' + meta_text + '}'
        metadata = json.loads(meta_json)
        
    except:
        metadata = None
    
    try:
        entry_citation = soup.find(attrs={'name':'description'}).attrs['content'] # type: ignore
    except:
        entry_citation = None
    
    if (metadata == None) and (entry_citation == None):
        return
    
    if entry_citation != None:
        
        description = entry_citation
        
        try:
            authors = entry_citation.split(metadata['itemTitle'])[0].strip().split(', ') # type: ignore
            authors = [i.strip(',') for i in authors]
        except:
            authors = None
        
        try:
            date_items = ' '.join([i for i in entry_citation.split(', ') if any(char.isdigit() for char in i)])
            date_items = date_items.replace('Vol. ', '').replace('No. ', '').replace('pp. ', '').replace('p. ', '').replace('-', ' ').replace('/', ' ').replace(')', '').replace('(', '').split(' ')
            date = [i for i in date_items if ((i.isdigit() == True) and (len(i) > 3))][0]
        except:
            date = None
            
    else:
        authors = None
        date = None
    
    try:
        link = soup.find(attrs={'rel':'canonical'}).attrs['href'] # type: ignore
    except:
        link = link
    
    
    if metadata != None:
        
        try:
            item_type = metadata['itemType']
        except:
            item_type = None
    
        try:
            doi = metadata['objectDOI']
        except:
            doi = None
    
        try:
            source = metadata['contentName']
        except:
            source = None
    
        try:
            publisher = metadata['contentPublisher']
        except:
            publisher = None
    
        try:
            title = metadata['itemTitle']
        except:
            title = None
    
    else:
        item_type = None
        doi = None
        source = None
        publisher = None
        title = None
    
    global results_cols
    result = pd.DataFrame(columns = results_cols, dtype=object)

    result.loc[0, 'title'] = title
    result.loc[0, 'authors'] = authors
    result.loc[0, 'date'] = date
    result.loc[0, 'source'] = source
    result.loc[0, 'publisher'] = publisher
    result.loc[0, 'type'] = item_type
    result.loc[0, 'description'] = description
    result.loc[0, 'repository'] = repository
    result.loc[0, 'doi'] = doi
    result.loc[0, 'link'] = link
    
    return result

def scrape_jstor(url = 'request_input'):
    
    if url == 'request_input':
        url = input('URL: ')
    
    if type(url) !=str:
        raise TypeError('URL must be a string')
    
    if ('jstor.org' not in url) and ('doi.org' not in url):
        raise ValueError('URL must be for a JSTOR webpage')
    
    try:
        res = scrape_url(url = url, parse_pdf = False, output = 'html')
    except:
        res = ''

    result = parse_jstor_from_source(source = res, link = url) # type: ignore
    
    return result

def parse_google_scholar_source(source = 'request_input'):
    
    if source == 'request_input':
        source = '"""' + input('Source code: ') + '"""'
    
    if type(source) != str:
        raise TypeError('Query must be a string')
    
    try:
        soup=BeautifulSoup(source,'lxml')
    except:
        soup = BeautifulSoup()
    
    try:
        data_lids = soup.select('[data-lid]')
    except:
        data_lids = []
    
    global results_cols
    df = pd.DataFrame(columns = results_cols, dtype=object)

    for item in data_lids:

        index = len(df.index)
        
        try:
            df.loc[index, 'title'] = item.select('h3')[0].get_text()
        except:
            df.loc[index, 'title'] = None
        try:
            df.loc[index, 'link'] = item.select('a')[0]['href']
        except:
            df.loc[index, 'link'] = None
        try:
            df.loc[index, 'extract'] = item.select('.gs_rs')[0].get_text()
        except:
            df.loc[index, 'extract'] = None
            
        author_list = []
        try:
            selected = item.select('a')
            
            for entry in selected:
                
                if 'related:' in entry['href']:
                    df.loc[index, 'recommendations'] = 'https://scholar.google.com/' + entry['href'] # type: ignore
            
                if 'user=' in entry['href']:
                    df.loc[index, 'author_link'] = 'https://scholar.google.com' + entry['href'] # type: ignore
                
                if 'cites=' in entry['href']:
                    df.loc[index, 'cited_by'] = 'https://scholar.google.com' + entry['href'] # type: ignore
            
                entry_text = entry.get_text()
                
                if (
                    ('[' not in entry_text)
                    and (']' not in entry_text)
                    and (']' not in entry_text)
                    and ('.c' not in entry_text)
                    and ('.n' not in entry_text)
                    and (entry_text not in df.loc[index, 'title']) # type: ignore
                    and (entry_text != 'Save') 
                    and (entry_text != 'Cite') 
                    and ('Cited by' not in entry_text) 
                    and (entry_text != 'Related articles') 
                    and ('versions' not in entry_text)
                    and (entry_text != 'View as HTML')
                    and (entry_text != '\n')
                    and (entry_text != 'Library Search')
                    ):
                        author_list.append(entry_text)
        except:
            pass
        
        df.loc[index, 'authors'] = author_list
            
        if ((df.loc[index, 'title'] == np.nan) or (df.loc[index, 'title'] == None)) and ('books.google' in df.loc[index, 'link']): # type: ignore

            try:
                link = df.loc[index, 'link']
                gbooks_response = scrape_url(url = link) # type: ignore
                title=BeautifulSoup(gbooks_response.content,'html').select('title')[0].get_text() # type: ignore
                df.loc[index, 'title'] = title
            except:
                df.loc[index, 'title'] = None
                
            

    df = df.replace(np.nan, None)
    df['title'] = df['title'].str.replace('[PDF]', '').str.replace('[BOOK]', '').str.replace('[B]', '').str.replace('[HTML]', '')
    df['authors'] = df['authors'].apply(join_list_by_colon).str.replace('cached', '').str.replace('full View', '')
    df['authors'] = df['authors'].apply(split_str_by_colon)
    
    return df

def search_google_scholar(query = 'request_input', pages = 1, open_source = False):
    
    if query == None:
        query = input('Query: ')
    query = urllib.parse.quote_plus(query) # type: ignore
    
    url_base = 'https://scholar.google.com/scholar?start='
    if open_source == True:
          url_base = 'view-source:' + url_base

    for i in range(0, pages):
            url = url_base + str(i)+ '0' + '&q=' + query
            open_url(url)

def open_google_scholar_links(source = None):
    
    if source == None:
        source = '"""' + input('Source code: ') + '"""'

    res = parse_google_scholar_source(source = source)

    for i in res['Link'].values:
        open_url(i)

def scrape_google_scholar(url):
    
    if type(url) != str:
        raise TypeError('Query must be a string')
    
    headers = {'User-Agent':'Mozilla/5.0 (Macintosh; Intel Mac OS X 10 _11_2) AppleWebkit/601.3.9 (KHTML, like Gecko) Version/9.0.2 Safari/601.3.9'}
   
    try:
        response=requests.get(url,headers=headers) 
        source = response.content
    except:
        raise ValueError('The scraper encountered an error. Google Scholar may have blocked it.')
    
    try:
        return parse_google_scholar_source(source = source) # type: ignore
    except:
        raise ValueError('The scraper encountered an error. Google Scholar may have blocked it.')

def scrape_google_scholar_search(query):
    
    if type(query) != str:
        raise TypeError('Query must be a string')
    
    
    url_base = 'https://scholar.google.com/scholar?q='
    query = urllib.parse.quote_plus(query) # type: ignore
    url = url_base + query

    return scrape_google_scholar(url)

def iterate_scholar_pages(scholar_page, page_limit = 20):
    
    global results_cols
    df = pd.DataFrame(columns = results_cols, dtype=object)

    url_end = scholar_page.split('?')[-1]
    
    if 'start=' not in url_end:

        for i in range(0, page_limit):
            page = str(i)+'0'
            new_url = 'https://scholar.google.com/scholar?start=' + page + '&' + url_end
            df = pd.concat([df, scrape_google_scholar(new_url)])   
            

    else:
        for i in range(0, page_limit):
            page = str(i)+'0'
            url_split = scholar_page.split('start=')
            url_start = url_split[0]
            url_end = url_split[1][2:]

            new_url = url_start + 'start=' + str(page) + url_end
            df = pd.concat([df, scrape_google_scholar(new_url)])
    

    df = df.reset_index().drop('index', axis=1)
    
    return df

def scrape_doi(doi):

    if doi.startswith('www.doi.org/') ==  True:
        doi = doi.replace('www.doi.org/', 'https://doi.org/')

    if doi.startswith('https://doi.org/') == False:
        doi = 'https://doi.org/' + doi
    
    if check_bad_url(doi) == False:

        final_url = get_final_url(doi)

        return scrape_article(url=final_url)
    else:
        raise ValueError('Bad DOI code or URL. Please check.')

def scrape_article(url = 'request_input') -> pd.DataFrame:

    if url == 'request_input':
            url = input('URL: ')

    if type(url) != str:
        raise TypeError('URL must be a string.')

    global results_cols
    df = pd.DataFrame(columns = results_cols)

    if ('doi.org' in url):
        return scrape_doi(url)

    if ('frontiersin.org' in url):
        return scrape_frontiers(url)
    
    if ('arxiv.org' in url):
        return scrape_arxiv(url)
    
    if ('springer' in url):
        return scrape_springer(url)
    
    if ('nature.com' in url):
        return scrape_nature(url)
    
    if ('ieee.org' in url):
        return scrape_ieee(url)
    
    if ('pubmed.ncbi.' in url):
        return scrape_pubmed(url)
    
    if ('gov/pmc' in url):
        return scrape_pmc(url)
    
    if ('ssrn.com' in url):
        return scrape_ssrn(url)
    
    if ('heinonline.org' in url):
        return scrape_heinonline(url)
    
    if ('mdpi.com' in url):
        return scrape_mdpi(url)
    
    if ('acm.org' in url):
        return scrape_acm(url)
    
    if ('muse.jhu.edu' in url):
        return scrape_muse(url)
    
    if ('proquest.com' in url):
        return scrape_proquest(url)
    
    if ('jstor.org' in url):
        return scrape_jstor(url)
    
    if ('scholar.google.com' in url):
        return scrape_google_scholar(url)

    else:
        return df
    


# Automated scraping not functional for CUP, OUP, SAGE, T&F, Science, Wiley, SciDirect, or ResearchGate sites due to bot blockers. 
## See notebook for scripts for parsing CUP, OUP, SAGE, T&F, Science, Wiley, SciDirect, and ResearchGate


from .network_exporters import export_network

import os
import copy
import json
import pickle
from random import Random

import pandas as pd
import numpy as np
from igraph import Graph # type: ignore
from networkx.classes import Graph as NetworkX_Undir, DiGraph as NetworkX_Dir, MultiGraph as NetworkX_Multi # type: ignore
from docx import Document # type: ignore

def export_obj(obj, file_name = 'obj_name', folder_address: str = 'request_input', export_str_as: str = 'txt', export_dict_as: str = 'json', export_pandas_as: str = 'csv', export_network_as: str = 'graphML'):
    
    """
    Exports objects to external files based on their type.
    """
    
    # Checking object type
    obj_type = type(obj)
    
    # Getting file name from user input
    if (file_name is None) or (file_name is np.nan):
        file_name = input('File name: ')
    
    # If file name is still None or is an empty string, assigns a series of random integers as file name
    if (file_name == '') or (file_name is None) or (file_name is np.nan):
        file_name = str(Random().randint(1000, 9999)) + '_' + str(Random().randint(1000, 9999)) + '_' + str(Random().randint(1000, 9999))
        file_name = str(file_name)
    
    # Getting folder address from user input
    if folder_address == 'request_input':
        folder_address = input('Folder address: ')
    
    # Creating file address
    file_address = folder_address + '/' + file_name
    
    # Converting list, set, tuple, and numeric types to strings
    if (obj_type == list) or (obj_type == set) or (obj_type == tuple) or (obj_type == int) or (obj_type == float):
        obj = str(obj)
        obj_type = type(obj)
    
    # If object is a string, exporting object as a text file
    if obj_type == str:
        
        # Default: exports file as .txt format
        if export_str_as.lower().strip() == 'txt':
        
            file_address = file_address + '.txt'
            with open(file_address, 'w', encoding='utf-8') as file:
                file.write(obj)
                file.close()

            return
        
        # Exporting file based on user input
        if (export_str_as.lower().strip() == 'word') or (export_str_as.lower().strip() == 'docx') or (export_str_as.lower().strip() == '.docx'):
            
            file_address = file_address + '.docx'
            document = Document()
            document.add_paragraph(obj)
            document.save(file_address)
            
            return
    
    # If object is a dictionary, exporting object as a JSON file
    if obj_type == dict:
        
        file_address = file_address + '.json'
        obj_copy = copy.deepcopy(obj)
        
        for key in obj.keys(): # type: ignore
            value = obj_copy[key]
            if (type(value) == dict) or (type(value) == pd.DataFrame) or (type(value) == pd.Series):
                str_version = str(value)
                del obj_copy[key] # type: ignore
                obj_copy[key] = str_version # type: ignore
        
        with open(file_address, "w") as file:
            json.dump(obj_copy, file)
            
        return
    
    # If object is a dateframe, exporting object as spreadsheet
    if (obj_type == pd.DataFrame) or (obj_type == pd.Series):
        
        # If export format selected is CSV, exporting .csv
        if (export_pandas_as.lower().strip() == 'csv') or (export_pandas_as.lower().strip() == '.csv'):
            file_address = file_address + '.csv'
            return obj.to_csv(file_address) # type: ignore
        
        # If export format selected is Excel, exporting .xlsx
        if (export_pandas_as.lower().strip() == 'excel') or (export_pandas_as.lower().strip() == 'xlsx')or (export_pandas_as.lower().strip() == '.xlsx'):
            file_address = file_address + '.xlsx'
            return obj.to_excel(file_address) # type: ignore
    
    # If object is a network, exporting object as a graph object
    if (obj_type == Graph) or (obj_type == NetworkX_Undir) or (obj_type==NetworkX_Dir) or (obj_type==NetworkX_Multi) or ('Network' in str(obj_type)):
        return export_network(obj, file_name = file_name, folder_address = folder_address, file_type = export_network_as)
    
    
    # For all other data types, exporting object as a .txt file based on its string representation
    else:
        obj = str(obj)
        file_address = file_address + '.txt'
        with open(file_address, 'w', encoding='utf-8') as file:
            file.write(obj)
            file.close()
        
        return

def art_class_to_folder(obj, folder_name = 'request_input', folder_address: str = 'request_input', export_str_as: str = 'txt', export_dict_as: str = 'json', export_pandas_as: str = 'csv', export_network_as: str = 'graphML'):

    obj_type_str = str(type(obj))

    # If the object is None, no folder created
    if obj is None:
        return
    
    # Getting folder name from user input
    if (folder_name is None) or (folder_name is np.nan) or (folder_name =='request_input'):
        folder_name = input('Folder name: ')
        

    # If folder name is still None or is an empty string, assigns a random integer as a folder name
    if (folder_name == '') or (folder_name == None):
        folder_name = str(Random().randint(1000, 9999)) + '_' + str(Random().randint(1000, 9999)) + '_' + str(Random().randint(1000, 9999))
        folder_name = str(folder_name)
    
     # Getting folder address from user input
    if folder_address == 'request_input':
        folder_address = input('Folder address: ')
    
    # Creating folder address
    folder_name = folder_name.strip().replace('/', '_').replace(' ', '_').replace('A:','').replace('F:','').replace('AUTH:','').strip()
    obj_address = str(folder_address) + '/' + str(folder_name)
    obj_address = obj_address.strip()

    if (('.Review' not in obj_type_str)
                        and ('.Results' not in obj_type_str)
                        and ('.ActivityLog' not in obj_type_str)
                        and ('.Entity' not in obj_type_str)
                        and ('.Entities' not in obj_type_str)
                        and ('.Author' not in obj_type_str)
                        and ('.Authors' not in obj_type_str)
                        and ('.Funder' not in obj_type_str)
                        and ('.Funders' not in obj_type_str)
                        and ('.Affiliation' not in obj_type_str)
                        and ('.Affiliations' not in obj_type_str)
                        and ('.Networks') not in obj_type_str):
        return

    if (('.Results' in obj_type_str)
            or ('.ActivityLog' in obj_type_str)):
        obj_address = obj_address + '.csv'
        obj.to_csv(obj_address)
        return

    if (('.Review' in obj_type_str)
                        or ('.Entity' in obj_type_str)
                        or ('.Entities' in obj_type_str)
                        or ('.Author' in obj_type_str)
                        or ('.Authors' in obj_type_str)
                        or ('.Funder' in obj_type_str)
                        or ('.Funders' in obj_type_str)
                        or ('.Affiliation' in obj_type_str)
                        or ('.Affiliations' in obj_type_str)
                        or ('.Networks' in obj_type_str)):

        os.mkdir(obj_address)

        for key in obj.__dict__.keys():
            
            attr = obj.__dict__[key]
            attr_name = str(key).replace('A:','').replace('F:','').replace('AUTH:','')

            if (attr is not None) and (attr_name != 'data'):
                
                attr_str_type = str(type(attr))
                
                if (('.Review' in attr_str_type)
                        or ('.Results' in attr_str_type)
                        or ('.ActivityLog' in attr_str_type)
                        or ('.Entity' in attr_str_type)
                        or ('.Entities' in attr_str_type)
                        or ('.Author' in attr_str_type)
                        or ('.Authors' in attr_str_type)
                        or ('.Funder' in attr_str_type)
                        or ('.Funders' in attr_str_type)
                        or ('.Affiliation' in attr_str_type)
                        or ('.Affiliations' in attr_str_type)
                        or ('.Networks' in attr_str_type)):
                    
                    art_class_to_folder(obj=attr,
                                        folder_name = attr_name, 
                                        folder_address= obj_address, 
                                        export_str_as = export_str_as, 
                                        export_dict_as = export_dict_as,
                                        export_pandas_as = export_pandas_as,
                                        export_network_as=export_network_as)
                    
                    continue

                else:

                    if (type(obj) == list) and (len(obj)>0) and ((type(obj[0]) == dict) or (type(obj[0]) == str) or (type(obj[0]) == dict) or (type(obj[0]) == set)or (type(obj[0]) == tuple)):
                        export_obj(obj, file_name = attr_name, folder_address = obj_address, export_str_as = export_str_as, export_dict_as = export_dict_as, export_pandas_as = export_pandas_as, export_network_as = export_network_as)
                        continue
                    
                    if (type(obj) == dict) and (len(obj.values())>0) and ((type(list(obj.values())[0]) == dict) or (type(list(obj.values())[0]) == list) or (type(list(obj.values())[0]) == str) or (type(list(obj.values())[0]) == set) or (type(list(obj.values())[0]) == tuple)):
                        export_obj(obj, file_name = attr_name, folder_address = obj_address, export_str_as = export_str_as, export_dict_as = export_dict_as, export_pandas_as = export_pandas_as, export_network_as = export_network_as)
                        continue
                    
                    else:
                    
                        obj_to_folder(obj=attr,
                                            folder_name = attr_name, 
                                            folder_address= obj_address, 
                                            export_str_as = export_str_as, 
                                            export_dict_as = export_dict_as,
                                            export_pandas_as = export_pandas_as,
                                            export_network_as=export_network_as)
                        
                        continue
            
    return

def obj_to_folder(obj, folder_name = 'request_input', folder_address: str = 'request_input', export_str_as: str = 'txt', export_dict_as: str = 'json', export_pandas_as: str = 'csv', export_network_as: str = 'graphML'):
    
    """
    Exports objects as external folders.
    """
    
    obj_type = type(obj)
    obj_type_str = str(type(obj))

    # If the object is None, no folder created
    if obj is None:
        return
    
    # Getting folder name from user input
    if (folder_name is None) or (folder_name is np.nan) or (folder_name =='request_input'):
        folder_name = input('Folder name: ')
        

    # If folder name is still None or is an empty string, assigns a random integer as a folder name
    if (folder_name == '') or (folder_name == None):
        folder_name = str(Random().randint(1000, 9999)) + '_' + str(Random().randint(1000, 9999)) + '_' + str(Random().randint(1000, 9999))
        folder_name = str(folder_name)
    
     # Getting folder address from user input
    if folder_address == 'request_input':
        folder_address = input('Folder address: ')
    
    # Creating folder address
    folder_name = folder_name.strip().replace('/', '_').replace(' ', '_')
    obj_address = str(folder_address) + '/' + str(folder_name)
    obj_address = obj_address.strip()
    
    
    # If the item is a custom ART class, recursively creates a folder using the art_class_to_folder() function.
    if (('.Review' in obj_type_str)
                        or ('.Results' in obj_type_str)
                        or ('.ActivityLog' in obj_type_str)
                        or ('.Entity' in obj_type_str)
                        or ('.Entities' in obj_type_str)
                        or ('.Author' in obj_type_str)
                        or ('.Authors' in obj_type_str)
                        or ('.Funder' in obj_type_str)
                        or ('.Funders' in obj_type_str)
                        or ('.Affiliation' in obj_type_str)
                        or ('.Affiliations' in obj_type_str)
                        or ('.Networks' in obj_type_str)):
        try:
            art_class_to_folder(obj = obj, folder_name = folder_name, folder_address = folder_address, export_str_as = export_str_as, export_dict_as = export_dict_as, export_pandas_as = export_pandas_as, export_network_as = export_network_as)
            return
        
        # Error handling
        except Exception as e:
            raise e

    # If the item is a string, numeric, Graph, Network, Pandas series or pandas.DataFrame, creates a file rather than a folder
    if (obj_type == str) or (obj_type == int) or (obj_type == float) or (obj_type == pd.Series) or (obj_type == pd.DataFrame) or (obj_type == Graph) or (obj_type == NetworkX_Undir) or (obj_type==NetworkX_Dir) or (obj_type==NetworkX_Multi) or ('.Network' in obj_type_str):
        folder_name = folder_name.strip().replace(' ', '_').replace('/', '_')
        export_obj(obj, file_name = folder_name, folder_address = folder_address, export_str_as = export_str_as, export_dict_as = export_dict_as, export_pandas_as = export_pandas_as, export_network_as = export_network_as)
        return
    
    

    # If the object is iterable, creates a folder using recursion
    if (obj_type == list) or (obj_type == set) or (obj_type == tuple):

        if len(obj) >0:

            # Creating folder
            os.mkdir(obj_address)

            index = 0
            for i in obj:
                item_folder_name = str(folder_name) + '_' + str(index)
                item_folder_name = item_folder_name.strip().replace(' ', '_').replace('/', '_')
                obj_to_folder(obj = i, folder_name = item_folder_name, folder_address = obj_address)
                index += 1
        
        return
    
    

    # If the object is a dictionary, uses recursion to create a folder with keys as filenames and values as files
    if obj_type == dict:

        if len(obj.keys()) > 0:
        
            # Creating folder
            os.mkdir(obj_address)

            for key in obj.keys():
                
                key_folder_name = str(folder_name) + '_' + str(key)
                obj_to_folder(obj = obj[key], folder_name = key_folder_name, folder_address = obj_address)
            
        return


from ..utils.basics import results_cols
from ..importers.crossref import lookup_doi
from ..internet.scrapers import get_final_url, scrape_url, scrape_article, can_scrape, get_domain, scrape_google_scholar
from ..internet.crawlers import check_crawl_permission

from ..internet.crawlers import correct_seed_errors as correct_seed_url_errors
from .references import extract_references

import queue
import time

import pandas as pd
import numpy as np


def crawler_scrape_url(url) -> pd.DataFrame:

    scrape_res = scrape_url(url=url)

    global results_cols
    result = pd.DataFrame(index = [0], columns=results_cols, dtype=object)

    keys = scrape_res.keys() # type: ignore

    if 'title' in keys:
        result.loc[0, 'title'] = scrape_res['title'] # type: ignore
    
    if 'author' in keys:
        auths = scrape_res['author'] # type: ignore
        if type(auths) == str:
            auths = auths.replace('[','').replace(']','').replace('{','').replace('}','').replace(';','').replace(', ',',').replace(' ,',',').split(',')
        result.at[0, 'authors'] = auths
        result.at[0, 'authors_data'] = auths

    if 'date' in keys:
        result.loc[0, 'date'] = scrape_res['date'] # type: ignore
    
    if ('sitename' in keys) and (scrape_res['sitename'] is not None) and (scrape_res['sitename'] != ''):  # type: ignore
            result.loc[0, 'source'] = scrape_res['sitename'] # type: ignore
    else:
        if ('url' in keys) and (scrape_res['url'] is not None) and (scrape_res['url'] != ''):  # type: ignore
            domain = get_domain(scrape_res['url']) # type: ignore
            result.loc[0, 'source'] = domain
    
    if ('type' in keys):
        result.loc[0, 'type'] = scrape_res['type'] # type: ignore
    else:
        if ('pagetype' in keys) and (scrape_res['pagetype'] is not None) and (scrape_res['pagetype'] != ''):  # type: ignore
            result.loc[0, 'type'] = scrape_res['pagetype'] # type: ignore
        else:
            result.loc[0, 'type'] = 'website'
    
    if ('publisher' in keys) and (scrape_res['publisher'] is not None) and (scrape_res['publisher'] != ''):  # type: ignore
        result.loc[0, 'publisher'] = scrape_res['publisher'] # type: ignore
    else:
        if ('sitename' in keys) and (scrape_res['sitename'] is not None) and (scrape_res['sitename'] != ''):  # type: ignore
            result.loc[0, 'publisher'] = scrape_res['sitename'] # type: ignore
        else:
            if 'url' in keys:
                domain = get_domain(scrape_res['url']) # type: ignore
                result.loc[0, 'publisher'] = domain
            else:
                result.loc[0, 'publisher'] = None
    
    if ('tags' in keys) and (scrape_res['tags'] is not None) and (scrape_res['tags'] != ''):  # type: ignore
        tags = scrape_res['tags'] # type: ignore
        tags_list = tags.replace('[','').replace(']','').replace('{','').replace('}','').replace(';','').replace(', ',',').replace(' ,',',').split(',')
        result.at[0, 'keywords'] = tags_list
    else:
        if ('categories' in keys) and (scrape_res['categories'] is not None) and (scrape_res['categories'] != ''):  # type: ignore
            cats = scrape_res['categories'] # type: ignore
            cats_list = cats.replace('[','').replace(']','').replace('{','').replace('}','').replace(';','').replace(', ',',').replace(' ,',',').split(',')
            result.at[0, 'keywords'] = cats_list
        else:
            result.at[0, 'keywords'] = []
    
    if ('description' in keys) and (scrape_res['description'] is not None) and (scrape_res['description'] != ''):  # type: ignore
        result.at[0, 'description'] = scrape_res['description'] # type: ignore

    if ('excerpt' in keys) and (scrape_res['excerpt'] is not None) and (scrape_res['excerpt'] != ''):  # type: ignore
        result.at[0, 'extract'] = scrape_res['excerpt'] # type: ignore
    
    if ('text' in keys) and (scrape_res['text'] is not None) and (scrape_res['text'] != ''):  # type: ignore
        result.at[0, 'full_text'] = scrape_res['text'] # type: ignore
    else:
        if ('raw_text' in keys) and (scrape_res['raw_text'] is not None) and (scrape_res['raw_text'] != ''):  # type: ignore
            result.at[0, 'full_text'] = scrape_res['raw_text'] # type: ignore
        else:
            if ('html' in keys) and (scrape_res['html'] is not None) and (scrape_res['html'] != ''):  # type: ignore
                result.at[0, 'full_text'] = scrape_res['html'] # type: ignore
    
    if ('links' in keys) and (scrape_res['links'] is not None) and (scrape_res['links'] != '') and (scrape_res['links'] != []): # type: ignore

        citations = scrape_res['links'] # type: ignore

        if type(citations) == str:
            citations = citations.replace('[','').replace(']','').replace('{','').replace('}','').replace(';','').replace(', ',',').replace(' ,',',').split(',')
        
        result.at[0, 'citations'] = citations
        result.at[0, 'citations_data'] = citations
        result.loc[0, 'citation_count'] = len(citations)
    
    if ('language' in keys) and (scrape_res['language'] is not None) and (scrape_res['language'] != ''):  # type: ignore
        result.at[0, 'language'] = scrape_res['language'] # type: ignore

    if ('url' in keys) and (scrape_res['url'] is not None) and (scrape_res['url'] != ''):  # type: ignore
        result.loc[0, 'link'] = scrape_res['url'] # type: ignore
    
    return result

def citation_crawler_site_test(url: str):

    global can_scrape

    for i in can_scrape:
        if i in url:
            return True
    
    return False

def academic_scraper(url, be_polite = False):

    # Checking if URL is bad. If True, tries to correct it.
    url = correct_seed_url_errors(url)
    domain = get_domain(url)
        
    # If be_polite is True, checks if crawler has permission to crawl/scrape URL
    if be_polite == True:
        if domain != 'acm.org': # ACM blanket refuses crawler permissions
            try:
                # If the crawler does not have permission, skips URL
                if check_crawl_permission(url) == False:
                    return pd.DataFrame(columns=results_cols)
            except:
                pass
    
    if 'scholar.google.com' in url:

        try:
            res_df = scrape_google_scholar(url)
        except:
            res_df = pd.DataFrame(columns=results_cols)

    else:
        if citation_crawler_site_test(url) == True:

            try:
                res_df = scrape_article(url)
            except:
                res_df = pd.DataFrame(columns=results_cols)

        else:
            try:
                res_df = crawler_scrape_url(url)
            except:
                res_df = pd.DataFrame(columns=results_cols)
    
    return res_df

def citation_crawler_scraper(entry: pd.Series, be_polite = True):
    
    url = entry['link']

    res_df = academic_scraper(url=url, be_polite=be_polite)
    
    if len(res_df) > 0: 
                res_series = res_df.loc[0]
                for i in res_series.index:
                    entry[i] = res_series[i]
    
    return entry
        
def citation_crawler_doi_retriver(entry: pd.Series, be_polite = True, timeout = 60):

    doi = entry['doi']
    link = entry['link']

    if type(doi) == str:
        try:
            res_df = lookup_doi(doi, timeout=timeout)

            if len(res_df) > 0:
                res_series = res_df.loc[0]
                for i in res_series.index:
                    entry[i] = res_series[i]
            
            return entry
    
        except:
                doi = doi.replace('https://', '').replace('http://', '').replace('dx.', '').replace('doi.org/', '')
                doi = 'https://doi.org' + doi
                
                try:
                    entry['link'] = doi
                    return citation_crawler_scraper(entry, be_polite = be_polite)
                except:

                    try:
                        entry['link'] = link
                        return citation_crawler_scraper(entry, be_polite = be_polite)
                    except:

                        try:
                            return crawler_scrape_url(doi)
                        except:
                            return entry
        
    else:
        if link != None:
            try:
                    return citation_crawler_scraper(entry, be_polite = be_polite)
            except:
                    try:
                        return crawler_scrape_url(link)
                    except:
                        return entry
            
        else:
            return entry

def update_citation_crawler_data(entry: pd.Series, be_polite = True, timeout = 60):

    doi = entry['doi']
    link = entry['link']

    if (doi != None) and (doi != 'None') and (doi != ''):
        try:
            return citation_crawler_doi_retriver(entry, be_polite = be_polite, timeout = timeout)
        except:
            return citation_crawler_scraper(entry, be_polite = be_polite)

    else:
        if (link != None) and (type(link) == str) and ('doi.org' in link):
            doi = link.replace('https://', '').replace('http://', '').replace('dx.', '').replace('doi.org/', '')
            return citation_crawler_doi_retriver(entry, be_polite = be_polite, timeout = timeout)
        
        else:
            if link != None:
                
                if type(link) == dict:
                    link = list(link.values())

                if type(link) == list:
                    link = link[0]
                
                if type(link) == str:
                    return citation_crawler_scraper(entry, be_polite = be_polite)
            
            else:
                return entry

def citation_crawler_engine(
                    to_crawl,
                    data: pd.DataFrame,
                    use_api: bool,
                    crawl_limit, 
                    depth_limit,
                    be_polite = True,
                    rate_limit = 0.05,
                    timeout = 60
                ):
    
    """
    Core functionality for citation crawler. Takes inputted review entries, parses their citations, retrieves data, adds to results, and repeats. and returns a dataframe of results.
    
    Parameters
    ---------- 
    urls : queue 
        ordered queue of URLs to be crawled.
    required_keywords : list 
        list of keywords which sites must contain to be crawled.
    excluded_keywords : list 
        list of keywords which sites must *not* contain to be crawled.
    excluded_url_terms : list 
        list of strings; link will be ignored if it contains any string in list.
    case_sensitive : bool 
        whether or not to ignore string characters' case.
    crawl_limit : int 
        how many URLs the crawler should visit before it stops.
    ignore_urls : list 
        list of URLs to ignore.
    ignore_domains : list 
        list of domains to ignore.
    be_polite : bool 
        whether to respect websites' permissions for crawlers.
    full : bool 
        whether to run a full scrape on each site. This takes longer.
    
    
    Returns
    -------
    output_dict : dict 
        a dictionary containing results from each crawled site.
    """
    
    # Intiailising variables to store the pages already visited
    data = data.copy(deep=True)
    
    crawled_entries = []
    iteration = 1
    
    added_in_cycle = 0
    processed_in_cycle = 0
    depth_marker = len(data)
    depth = 1

    # until all pages have been visited
    
    while (not to_crawl.empty() == True) and (depth <= depth_limit) and (len(crawled_entries) <= crawl_limit):
        
        # Checking if crawler has crawled the maximum number of entries. If so, halts.
        if len(crawled_entries) > crawl_limit:
            print('\nCrawl limit reached')
            break
            
        # Checking if crawler has reached the depth limit. If so, halts.
        if depth > depth_limit:
            print('\nDepth limit reached')
            break
        
        # Getting the entry index to process from the queue
        _, current_index = to_crawl.get()
        
        # Checking if entry index has already been processed. If True, skips.
        if current_index in crawled_entries:
            continue
        
        old_len = len(data)

        # Retreiving entry data
        entry = pd.Series(data.loc[current_index])
        
        if use_api == True:
        # Checking if the entry has a valid DOI. If yes, updating data in the entry using Crossref API. If not, updating data using web scraper.
            entry = update_citation_crawler_data(entry, be_polite = be_polite, timeout = timeout)
            entry = pd.Series(entry)  # type: ignore


        # Formatting entry citations data
        refs = extract_references(entry['citations_data'], add_work_ids = True, update_from_doi = False)
        entry.at['citations'] = refs

        # Formatting entry authors data
        entry.at['authors'] = format_authors(entry['authors']) # type: ignore

        data.loc[current_index] = entry
        refs_df = refs.copy(deep=True) # type: ignore
        data = pd.concat([data, refs_df]).reset_index().drop('index', axis=1) # type: ignore

        # Adding current current index to list of indexes already processed
        crawled_entries.append(current_index)

        new_len = len(data)
        new_indexes = list(range(old_len, new_len))

        added_in_cycle = added_in_cycle + len(refs_df)


        for i in new_indexes:
            priority_score = 0.0001
            to_crawl.put((priority_score, i))
        
        
        # Pausing crawler for a brief interval to reduce server loads and avoid being blocked
        time.sleep(rate_limit) 
        
        # Displaying status to user

        if iteration > depth_marker:

            # print(f'\nCrawl depth {depth}:\n    - Entries processed: {processed_in_cycle}\n    - Results added: {added_in_cycle}')

            depth += 1
            added_in_cycle = 0
            processed_in_cycle = 0
            depth_marker = len(data)
            

        # Incrementing iteration count
        iteration += 1
        processed_in_cycle += 1

    # Updating newly added entries
    if use_api == True:
        index = set(data.index)
        not_crawled = list(index.difference(set(crawled_entries)))
        for i in not_crawled:
            entry = data.loc[i]
            entry = update_citation_crawler_data(entry, be_polite = be_polite, timeout = timeout)
            data.loc[i] = entry

    return data

def citation_crawler(
                    data: pd.DataFrame,
                    use_api: bool = True,
                    crawl_limit: int = 5, 
                    depth_limit: int = 2,
                    be_polite: bool = True,
                    rate_limit: float = 0.05,
                    timeout: int = 60
                    ) -> pd.DataFrame:
    
        """
        Crawls results, their citations, and so on.
        
        The crawler iterates through queue of works; extracts their citations; runs checks to validate each reference;
        based on these, selects a source to retrieve data from: (a) Crossref API (if has a valid DOI); (b) bespoke web scraping for 
        if so, scrapes them and adds links found to queue.
        
        Parameters
        ---------- 
        seeds : str or list 
            one or more URLs from which to crawl.
        crawl_limit : int 
            how many URLs the crawler should visit before it stops.
        excluded_url_terms : list 
            list of strings; link will be ignored if it contains any string in list.
        required_keywords : list 
            list of keywords which sites must contain to be crawled.
        excluded_keywords : list 
            list of keywords which sites must *not* contain to be crawled.
        case_sensitive : bool 
            whether or not to ignore string characters' case.
        ignore_urls : list 
            list of URLs to ignore.
        ignore_domains : list 
            list of domains to ignore.
        be_polite : bool 
            whether respect websites' permissions for crawlers.
        full : bool 
            whether to run a full scrape on each site. This takes longer.
        output_as : str 
            the format to output results in. Defaults to a pandas.DataFrame.
        
        
        Returns
        -------
        result : pd.DataFrame 
            an object containing the results of a crawl.
        """

        # See https://www.zenrows.com/blog/web-crawler-python#transitioning-to-a-real-world-web-crawler

        seeds = data.index.to_list()
        
        # Storing seed indexes to crawl in a specific order
        to_crawl = queue.PriorityQueue()
        
        # Queing seeds with highest priority
        for seed in seeds:
            to_crawl.put((0.0001, seed))
        
        # Running crawler engine
        output = citation_crawler_engine(
                    to_crawl,
                    data,
                    use_api,
                    crawl_limit, 
                    depth_limit,
                    be_polite,
                    rate_limit,
                    timeout
                )
        
        
        
        # Printing end status for user
        print('\n\n----------------------\nCrawl complete\n----------------------')
        
        return output




"""Functions to load and parse JSTOR database files"""

from ..utils.basics import results_cols

import webbrowser
from pathlib import Path

import pandas as pd
import numpy as np

def access_jstor_database():
    return webbrowser.open('https://constellate.org/builder')

def import_metadata(file_path = 'request_input',
                        #   clean_results = True
                          ):
    
    if file_path == 'request_input':
        file_path = input('File path: ')

    import_df = pd.read_csv(file_path, dtype=object).replace(np.nan, None)
    import_df = import_df.rename(columns=
                                        {'id': 'other_ids',
                                         'isPartOf': 'source',
                                        'datePublished': 'date',
                                        'docType': 'type',
                                        'provider': 'repository',
                                        'url': 'link',
                                        'placeOfPublication': 'publisher_location',
                                        'creator': 'authors',
                                        'keyphrase': 'keywords'
                                            })
    
    import_df = import_df.dropna(axis='columns', how='all')

    global results_cols

    to_drop = []
    for c in import_df.columns:
        if c not in results_cols:
            to_drop.append(c)
    


    output_df = import_df.drop(labels=to_drop, axis='columns')
    output_df['authors'] = output_df['authors'].str.lower().str.split(';')
    output_df['authors_data'] = output_df['authors'].copy(deep=True)
    output_df['keywords'] = output_df['keywords'].str.lower().str.split(';')

    output_df = output_df.replace(np.nan, None).replace('untitled', None)
    
    
    # if clean_results == True:
        
    #     cleaning_list = ['book review:',
    #                     '\[',
    #                     'references',
    #                      'index'
    #                     ]

    #     for item in cleaning_list:
    #         instances = output_df[output_df['title'].str.contains(item) == True].index.to_list()
    #         for i in instances:
    #             output_df.loc[i, 'title'] = None
    
    return output_df

def import_full(file_path = 'request_input', 
                # clean_results = True
                ):
    
    if file_path == 'request_input':
        file_path = input('File path: ')
    
    import_df = pd.read_json(file_path, lines=True).replace(np.nan, None)
    import_df = import_df.rename(columns=
                                        {'id': 'other_ids',
                                         'isPartOf': 'source',
                                        'datePublished': 'date',
                                        'docType': 'type',
                                        'provider': 'repository',
                                        'url': 'link',
                                        'placeOfPublication': 'publisher_location',
                                        'creator': 'authors',
                                        'keyphrase': 'keywords',
                                        'fullText': 'full_text'
                                            })
    
    import_df = import_df.dropna(axis='columns', how='all')

    global results_cols

    to_drop = []
    for c in import_df.columns:
        if c not in results_cols:
            to_drop.append(c)
    
    output_df = import_df.drop(labels=to_drop, axis='columns')
    output_df['authors'] = output_df['authors'].str.lower().str.split(';')
    output_df['authors_data'] = output_df['authors'].copy(deep=True)
    output_df['keywords'] = output_df['keywords'].str.lower().str.split(';')

    output_df = output_df.replace(np.nan, None).replace('untitled', None)
    
    
    # if clean_results == True:
        
    #     cleaning_list = ['book review:',
    #                     '\[',
    #                     'references',
    #                      'index'
    #                     ]

    #     for item in cleaning_list:
    #         instances = output_df[output_df['title'].str.contains(item) == True].index.to_list()
    #         for i in instances:
    #             output_df.loc[i, 'title'] = None
    
    return output_df

def import_jstor(file_path = 'request_input') -> pd.DataFrame:

    if file_path == 'request_input':
        file_path = input('File path: ')
    
    path_obj = Path(file_path)
    suffix = path_obj.suffix.strip('.')

    if (suffix != 'csv') and (suffix != 'json'):
        raise TypeError('File must be a CSV or JSON')

    if suffix == 'csv':
        result = import_metadata(file_path=file_path)
    
    if suffix == 'json':
        result = import_full(file_path=file_path)
    
    return result


from ..utils.basics import results_cols
from ..importers.crossref import references_to_df
from .results import Results

import pandas as pd
import numpy as np


class References(Results):

    """
    This is a References object. It is a modified Pandas Dataframe object designed to store References relating to an entry.
    
    Parameters
    ----------
    
    
    Attributes
    ----------
    """

    def __init__(self, dataframe = None, index = []):
        
        """
        Initialises References instance.
        
        Parameters
        ----------
        """

        
        # Inheriting methods and attributes from References class
        super().__init__()
            
        self.replace(np.nan, None)
    
    def __repr__(self):

        return f'References object containing {len(self)} items'

    def from_dataframe(dataframe): # type: ignore
        
        dataframe = dataframe.copy(deep=True).reset_index().drop('index', axis=1)
        results_table = References(index = dataframe.index)
        results_table.columns = results_table.columns.astype(str).str.lower().str.replace(' ', '_')
        dataframe.columns = dataframe.columns.astype(str).str.lower().str.replace(' ', '_')

        for c in dataframe.columns:
            results_table[c] = dataframe[c]
        
        return results_table



def is_formatted_reference(item):

    if type(item) == References:
        return True
    else:
        return False


def extract_references(references_data, add_work_ids = True, update_from_doi = False):

    refs = References()

    if type(references_data) == References:
        refs = references_data
        refs.refs_count = None

    
    if (type(references_data) != References) and (type(references_data) != str) and (type(references_data) != list) and (type(references_data) != dict):
        refs.refs_count = None
        return refs

    if (type(references_data) == list) and (len(references_data) > 0) and (type(references_data[0]) == dict):
        
        if 'count' in references_data[0].keys():
            count = references_data[0]['count']
        else:
            count = None

        df = references_to_df(references_data, update_from_doi = update_from_doi)
        df.replace({np.nan: None})
        refs = References.from_dataframe(df) # type: ignore
        refs.refs_count = count

        if 'db' in refs.columns:
            refs = refs.drop('db', axis=1)
        
        if 'count' in refs.columns:
            refs = refs.drop('count', axis=1)
        
        refs = refs.dropna(axis=0, how='all').reset_index().drop('index', axis=1)
        refs = References.from_dataframe(dataframe=refs) # type: ignore

        if add_work_ids == True:
            refs.generate_work_ids() # type: ignore
    
    if (type(references_data) == list) and (len(references_data) > 0) and (type(references_data[0]) == str):
        df = pd.DataFrame(columns=results_cols, dtype=object)
        df['link'] = pd.Series(references_data, dtype=object)
        df.replace({np.nan: None})

        refs = References.from_dataframe(df) # type: ignore
        if add_work_ids == True:
            refs.generate_work_ids()

        refs.refs_count = None


    return refs




from ..utils.basics import results_cols
from ..utils.cleaners import strip_list_str, deduplicate
from ..importers.pdf import read_pdf_to_table
from ..importers.jstor import import_jstor
from ..importers.bibtex import import_bibtex
from ..importers.crossref import lookup_doi, lookup_dois
from ..datasets import stopwords

from .entities import Entity, Entities
from .funders import Funder, Funders, format_funders

from pathlib import Path

import pandas as pd
import numpy as np
from nltk.tokenize import word_tokenize # type: ignore
from pybtex.database import BibliographyData, Entry #type: ignore

def generate_work_id(work_data: pd.Series):

        work_data = work_data.copy(deep=True).dropna()

        work_id = 'W:'
        
        if 'authors' in work_data.index:
            auths_type = type(work_data['authors'])
            auths_type_str = str(auths_type)

            if auths_type == list:
                work_data['authors'] = pd.Series(work_data['authors'],  dtype=object).sort_values().to_list()
            
            else:
                if '.Authors' in auths_type_str:
                    work_data['authors'] = work_data['authors'].summary['full_name'].sort_values().to_list()
            

        work_data = work_data.astype(str).str.lower()

        if 'authors' in work_data.index:
            authors = work_data['authors']
        else:
            authors = ''
        
        if 'title' in work_data.index:
            title = work_data['title']
        else:
            title = ''
        
        if 'date' in work_data.index:
            date = work_data['date']
        else:
            date = ''

        if (authors != None) and (authors != '') and (authors != 'None') and (authors != [])and (authors != '[]'):
            authors_str = authors.lower().strip().replace('[','').replace(']','').replace("'", "").replace('"', '')
            authors_list = authors_str.split(',')
            authors_list = [i.strip() for i in authors_list]
            
            if len(authors_list) > 0:
                authors_sorted = (pd.Series(authors_list, dtype=object).sort_values())
                first_author = str(authors_sorted[0])
                first_author = first_author.split(' ')[-1].split(' ')[-1]
            else:
                first_author = ''
            
            work_id = work_id + '-' + first_author

        if (title != None) and (title != '') and (title != 'None'):
            title = title.strip().lower()
            title_words = list(word_tokenize(title))
            title_first2 = '-'.join(title_words[:2])
            if len(title_words) > 3:
                title_last = title_words[-1]
            else:
                title_last = ''
            title_shortened = title_first2 + '-' + title_last
            title_shortened = title_shortened[:15] # capping at 15 characters to avoid overly long UIDs
            work_id = work_id + '-' + title_shortened
        
        if (date != None) and (date != '') and (date != 'None'):
            work_id = work_id + '-' + str(date)

        uid = ''

        if 'doi' in work_data.index:
            uid = work_data['doi']

            if (uid == None) or (uid == 'None') or (uid == ''):
                if 'isbn' in work_data.index:
                    uid = work_data['isbn']

                if (uid == None) or (uid == 'None') or (uid == ''):
                    if 'issn' in work_data.index:
                        uid = work_data['issn']

                    if (uid == None) or (uid == 'None') or (uid == ''):
                        if 'link' in work_data.index:
                            uid = work_data['link'][:15] # keeping URLs short as these can produce very long IDs

                        if (uid == None) or (uid == 'None') or (uid == ''):
                            uid = ''
        
        uid_shortened = uid.replace('https://', '').replace('http://', '').replace('www.', '').replace('doi.org.','').replace('scholar.google.com/','')[:23] # DOIs are 23 characters long

        work_id = work_id + '-' + uid_shortened
        work_id = work_id.replace('W:-', 'W:').replace("'s", '').replace('\r', '').replace('\n', '').replace("'", "").replace('"', '').replace('(','').replace(')','').replace('`','').replace('.', '').replace('’','').replace('--', '-').replace('W:-', 'W:').strip('-')
        work_id = work_id

        return work_id


class Results(pd.DataFrame):

    """
    This is a Results object. It is a modified Pandas Dataframe object designed to store the results of an academic review.
    
    Parameters
    ----------
    
    
    Attributes
    ----------
    """

    def __init__(self, dataframe = None, index = []):
        
        """
        Initialises Results instance.
        
        Parameters
        ----------
        """

        if dataframe is None:

            global results_cols

            # Inheriting methods and attributes from Pandas.DataFrame class
            super().__init__(dtype=object, columns = results_cols, index = index # type: ignore
                            ) 
            
            self.replace(np.nan, None)
        
        else:
            df = dataframe
            if type(dataframe) == pd.DataFrame:
                self = Results.from_dataframe(dataframe = df)

    def drop_empty_rows(self):

        ignore_cols = ['work_id', 'authors', 'funder', 'citations']

        df = self.dropna(axis=0, how='all')
        drop_cols = [c for c in df.columns if c not in ignore_cols]
        df = df.dropna(axis=0, how='all', subset=drop_cols).reset_index().drop('index', axis=1)

        results = Results.from_dataframe(dataframe=df, drop_duplicates=False) # type: ignore

        self.__dict__.update(results.__dict__)

        return self




    def remove_duplicates(self, drop_empty_rows = True, update_from_api = False):

        if drop_empty_rows == True:
            self.drop_empty_rows()

        self['doi'] = self['doi'].str.replace('https://', '', regex = False).str.replace('http://', '', regex = False).str.replace('dx.', '', regex = False).str.replace('doi.org/', '', regex = False).str.replace('doi/', '', regex = False)

        df = deduplicate(self)
        results = Results.from_dataframe(dataframe = df, drop_duplicates=False)

        if update_from_api == True:
            results.update_from_dois()
        
        results.update_work_ids()
        df2 = results.drop_duplicates(subset='work_id').reset_index().drop('index',axis=1)

        results2 = Results.from_dataframe(dataframe=df2, drop_duplicates=False) # type: ignore
        
        self.__dict__.update(results2.__dict__)

        return self

        

    def get(self, work_id: str):

        indexes = self[self['work_id'] == work_id].index.to_list()
        if len(indexes) > 0:
            index = indexes[0]
            return self.loc[index]
        else:
            raise KeyError('work_id not found')

    def add_pdf(self, path = 'request_input'):
        
        if path == 'request_input':
            path = input('Path to PDF (URL or filepath): ')

        table = read_pdf_to_table(path)
        table = table.replace(np.nan, None).astype(object)

        series = table.loc[0]
        work_id = generate_work_id(series) # type: ignore
        series['work_id'] = work_id

        index = len(self)
        self.loc[index] = series

        return self
    
    def add_row(self, data, drop_duplicates=True):

        if type(data) != pd.Series:
            raise TypeError(f'Results must be a Pandas.Series, not {type(data)}')

        data.index = data.index.astype(str).str.lower().str.replace(' ', '_')
        if len(data) != len(self.columns):
            for c in data.index:
                if c not in self.columns:
                    self[c] = pd.Series(dtype=object)

        index = len(self)

        work_id = generate_work_id(data)
        # work_id = self.get_unique_id(work_id, index)
        data['work_id'] = work_id

        
        self.loc[index] = data

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=False)
        
    def get_unique_id(self, work_id, index):

        if (type(work_id) == str) and (work_id != ''):
            try:
                work_id = str(work_id.split('#')[0])
                if work_id in self['work_id'].to_list():
                
                    df = self.copy(deep=True).astype(str)
                    df['work_id'] = df['work_id'].dropna()
                    masked = df[df['work_id'].str.contains(work_id)]
                    masked_indexes = masked.index.to_list()
                    if index not in masked_indexes:
                        id_count = len(masked) # type: ignore
                        work_id = work_id + f'#{id_count + 1}'
            except:
                pass
        return work_id

    def add_dataframe(self, dataframe: pd.DataFrame, drop_empty_rows = True, drop_duplicates = False, update_work_ids = True):
        
        if (type(dataframe) != pd.DataFrame) and (type(dataframe) != pd.Series):
            raise TypeError(f'Results must be a Pandas.Series or Pandas.DataFrame, not {type(dataframe)}')

        dataframe = dataframe.reset_index().drop('index', axis=1)
        dataframe.columns = dataframe.columns.astype(str).str.lower().str.replace(' ', '_')

        if (self.columns.to_list()) != (dataframe.columns.to_list()):
            for c in dataframe.columns:
                if c not in self.columns:
                    self[c] = pd.Series(dtype=object)
        
        self_copy = self.copy(deep=True)
        concat_df = pd.concat([self_copy, dataframe])
        concat_df = concat_df.reset_index().drop('index', axis=1)

        new_results = Results()

        for c in concat_df.columns:
            new_results[c] = concat_df[c]
        
        self.__dict__.update(new_results.__dict__)

        if drop_empty_rows == True:
            self.drop_empty_rows()

        if update_work_ids == True:
            self.update_work_ids(drop_duplicates=False)

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

    def add_doi(self, doi: str = 'request_input', drop_empty_rows = True, drop_duplicates = False, timeout: int = 60):
        df = lookup_doi(doi=doi, timeout=timeout)
        self.add_dataframe(dataframe=df)

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=False)
        
        if drop_empty_rows == True:
            self.drop_empty_rows()



    def add_dois(self, dois_list: list = [], drop_empty_rows = True, rate_limit: float = 0.1, timeout = 60):
        df = lookup_dois(dois_list=dois_list, rate_limit=rate_limit, timeout=timeout)
        self.add_dataframe(dataframe=df, drop_empty_rows = True)

    def correct_dois(self, drop_duplicates = False):

        no_doi = self[self['doi'].isna()]
        has_link = no_doi[~no_doi['link'].isna()]
        doi_in_link = has_link[has_link['link'].str.contains('doi.org')]

        for i in doi_in_link.index:
            link = str(doi_in_link.loc[i, 'link'])
            doi = link.replace('http://', '').replace('https://', '').replace('www.', '').replace('dx.', '').replace('doi.org/', '').strip('/').strip()
            self.loc[i, 'doi'] = doi
        
        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=False)

    def generate_work_ids(self):

        for i in self.index:
            work_id = generate_work_id(self.loc[i])
            self.loc[i, 'work_id'] = work_id

    def update_work_ids(self, drop_duplicates = False):

        for i in self.index:
            work_id = generate_work_id(self.loc[i])
            if self.loc[i, 'work_id'] != work_id:
                # work_id = self.get_unique_id(work_id, i)
                self.loc[i, 'work_id'] = work_id
        
        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=False)


    def update_from_doi(self, index, drop_empty_rows = True, drop_duplicates = False, timeout: int = 60):
        
        try:
            old_series = self.loc[index]
            doi = old_series['doi']

            if (type(doi) != None) and (type(doi) != '') and (type(doi) != 'None'):
                new_series = lookup_doi(doi=doi, timeout=timeout).loc[0]

                for i in new_series.index:
                    if i in old_series.index:
                        old_val = old_series[i]
                        new_val = new_series[i]
                        if (new_val != None) and (new_val != np.nan) and (new_val != '') and (len(str(new_val)) > len(str(old_val))):
                            old_series[i] = new_val
                    
                    else:
                        old_series[i] = new_val
        except:
            pass

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=False)
        
        if drop_empty_rows == True:
            self.drop_empty_rows()
        

    def update_from_dois(self, drop_empty_rows = True, drop_duplicates = False, timeout: int = 60):

        self.correct_dois(drop_duplicates=False)

        for i in self.index:
            self.update_from_doi(index = i, drop_duplicates = False, timeout=timeout)
        
        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=False)
        
        if drop_empty_rows == True:
            self.drop_empty_rows()

    def __add__(self, results_obj):

        left = self.copy(deep=True)
        right = results_obj.copy(deep=True)
        right.columns = right.columns.astype(str).str.lower().str.replace(' ', '_')

        if (left.columns.to_list()) != (right.columns.to_list()):
            for c in left.columns:
                if c not in right.columns:
                    right[c] = pd.Series(dtype=object)
            
            for c in right.columns:
                if c not in left.columns:
                    left[c] = pd.Series(dtype=object)
        
        index = len(left)
        for i in right.index:
            left.loc[index] = right.loc[i]
            index += 1
        
        return left

    def to_dataframe(self):
        return self.copy(deep=True)
    
    def from_dataframe(dataframe, drop_empty_rows = False, drop_duplicates = False): # type: ignore
        
        dataframe = dataframe.copy(deep=True).reset_index().drop('index', axis=1)
        results_table = Results(index = dataframe.index)
        results_table.columns = results_table.columns.astype(str).str.lower().str.replace(' ', '_')
        dataframe.columns = dataframe.columns.astype(str).str.lower().str.replace(' ', '_')

        for c in dataframe.columns:
            results_table[c] = dataframe[c]

        if drop_duplicates == True:
            results_table.remove_duplicates(drop_empty_rows=False)
        
        if drop_empty_rows == True:
            results_table.drop_empty_rows()

        return results_table

    def to_pybtex(self):

        res_dict = {}

        for i in self.index:
            
            row = self.loc[i].copy(deep=True).dropna()

            if 'type' in row.index:
                entry_type = row['type']
            else:
                entry_type = 'misc'
            
            if 'title' in row.index:
                title = str(row['title'])
            else:
                title = ''
            
            if 'authors' in row.index:
                authors = row['authors']
            else:
                authors = ''
            
            if ('__dict__' in authors.__dir__()) and ('summary' in authors.__dict__.keys()):
                if (type(authors.summary) == pd.DataFrame) and ('full_name' in authors.summary.columns):
                    authors_str = ', '.join(authors.summary['full_name'].to_list())
                else:
                    authors_str = ''
            
            if type(authors) == list:
                authors_str = ', '.join(authors)
            else:
                authors_str = ''
            
            if 'date' in row.index:
                year = str(row['date'])
            else:
                year = ''
            
            if 'keywords' in row.index:
                keywords = ', '.join(row['keywords'])
            else:
                keywords = ''

            if 'doi' in row.index:
                doi = str(row['doi'])
            else:
                doi = ''
            
            if 'publisher' in row.index:
                publisher = str(row['publisher'])
            else:
                publisher = ''
            
            if 'link' in row.index:
                link = str(row['link'])
            else:
                link = ''
            
            if 'isbn' in row.index:
                isbn = str(row['isbn'])
            else:
                isbn = ''

            if 'work_id' in row.index:
                key = row['work_id']
            else:
                key = str(title).lower() + '_' + str(authors_str)[:10].lower() + '_' + str(year).lower()

            entry_list = []

            if (authors_str is not None) and (authors_str != '') and (type(authors_str) == str):
                authors_tuple = ('author', authors_str)
                entry_list.append(authors_tuple)
            
            if (title is not None) and (title != '') and (type(title) == str):
                title_tuple = ('title', title)
                entry_list.append(title_tuple)
            
            if (year is not None) and (year != '') and (type(year) == str):
                year_tuple = ('year', year)
                entry_list.append(year_tuple)
            
            if (doi is not None) and (doi != '') and (type(doi) == str):
                doi_tuple = ('doi', doi)
                entry_list.append(doi_tuple)
            
            if (publisher is not None) and (publisher != '') and (type(publisher) == str):
                publisher_tuple = ('publisher', publisher)
                entry_list.append(publisher_tuple)
            
            if (link is not None) and (link != '') and (type(link) == str):
                link_tuple = ('url', link)
                entry_list.append(link_tuple)
            
            if (isbn is not None) and (isbn != '') and (type(isbn) == str):
                isbn_tuple = ('isbn', isbn)
                entry_list.append(isbn_tuple)
            
            if (keywords is not None) and (keywords != '') and (type(keywords) == str):
                keywords_tuple = ('keywords', keywords)
                entry_list.append(keywords_tuple)
            
            if 'article' in entry_type:

                entry_type = 'article'

                if 'source' in row.index:
                    journal = row['source']
                    if (journal is not None) and (journal != '') and (type(journal) == str):
                        journal_tuple = ('journal', journal)
                        entry_list.append(journal_tuple)

            if ('book' in entry_type) and ('chapter' in entry_type):

                entry_type = 'incollection'

                if 'source' in row.index:
                    booktitle = row['source']
                    if (booktitle is not None) and (booktitle != '') and (type(booktitle) == str):
                        booktitle_tuple = ('journal', booktitle)
                        entry_list.append(booktitle_tuple)
            
            if 'book' in entry_type:
                entry_type = 'book'

            entry = Entry(entry_type, entry_list)
            res_dict[key] = entry
        
        bib_data = BibliographyData(res_dict)

        return bib_data



    def to_bibtex(self):

        bib_data = self.to_pybtex()
        return bib_data.to_string('bibtex')
    
    def to_yaml(self):
        bib_data = self.to_pybtex()
        return bib_data.to_string('yaml')
    
    def export_bibtex(self, file_name = 'request_input', folder_path = 'request_input'):

        if file_name == 'request_input':
            file_name = input('File name: ')
        
        if folder_path == 'request_input':
            folder_path = input('Folder path: ')
        
        if Path(folder_path).exists() == False:
            raise ValueError('Folder does not exist')

        bib = self.to_bibtex()

        filepath = folder_path + '/' + file_name + '.bib'

        bib_bytes = bytes(bib, "utf-8").decode("unicode_escape")

        with open(filepath, 'w') as file:
            file.write(bib_bytes)
    

    def export_yaml(self, file_name = 'request_input', folder_path = 'request_input'):

        if file_name == 'request_input':
            file_name = input('File name: ')
        
        if folder_path == 'request_input':
            folder_path = input('Folder path: ')
        
        if Path(folder_path).exists() == False:
            raise ValueError('Folder does not exist')

        yaml = self.to_yaml()

        filepath = folder_path + '/' + file_name + '.yaml'

        yaml_bytes = bytes(yaml, "utf-8").decode("unicode_escape")

        with open(filepath, 'w') as file:
            file.write(yaml_bytes)
        
    def clear_rows(self):

        results = Results()
        self.__dict__.update(results.__dict__)

        return self

    def import_bibtex(self, file_path = 'request_input'):

        df = import_bibtex(file_path = file_path)
        self.add_dataframe(dataframe=df, drop_duplicates=False, drop_empty_rows=False)
    
    def from_bibtex(file_path = 'request_input'):

        results = Results()
        results.import_bibtex(file_path=file_path)

        return results


    def import_excel(self, file_path = 'request_input', sheet_name = None):

        if file_path == 'request_input':
            file_path = input('File path: ')

        if sheet_name == None:
            sheet_name = 0
        
        excel_import = pd.read_excel(file_path, sheet_name = sheet_name, header = 0, index_col = 0).replace({np.nan: None, 'none': None})
        
        self.add_dataframe(excel_import)

        cols = ['authors', 'keywords']
        for col in cols:
            for i in self.index:
                items = self.loc[i, col]
                
                try:
                    self.at[i, col] = strip_list_str(items) # type: ignore
                    
                except:
                    pass

        return self

    def from_excel(file_path = 'request_input', sheet_name = None): # type: ignore

        results_table = Results()
        results_table = results_table.import_excel(file_path, sheet_name).replace(np.nan, None) # type: ignore
        results_table.format_authors() # type: ignore


        return results_table

    def import_csv(self, file_path = 'request_input'):

            if file_path == 'request_input':
                file_path = input('File path: ')
                
            csv_import = pd.read_csv(file_path, header = 0, index_col = 0).replace({np.nan: None, 'none': None})
            self.add_dataframe(csv_import)

            cols = ['authors', 'keywords']
            for col in cols:
                for i in self.index:
                    items = self.loc[i, col]
                    
                    try:
                        self.at[i, col] = strip_list_str(items) # type: ignore
                        
                    except:
                        pass

            return self
    
    def from_csv(file_path = 'request_input'): # type: ignore

        results_table = Results()
        results_table.import_csv(file_path).replace(np.nan, None) # type: ignore


        return results_table

    def import_json(self, file_path = 'request_input'):

        if file_path == 'request_input':
                file_path = input('File path: ')

        json_import = pd.read_json(file_path)
        self.add_dataframe(json_import)

        cols = ['authors', 'keywords']
        for col in cols:
                for i in self.index:
                    items = self.loc[i, col]
                    
                    try:
                        self.at[i, col] = strip_list_str(items) # type: ignore
                        
                    except:
                        pass

        self = self.replace(np.nan, None)

        return self
    
    def from_json(file_path = 'request_input'): # type: ignore

        results_table = Results()
        results_table.import_json(file_path).replace(np.nan, None) # type: ignore
        
        return results_table
    
    def import_file(self, file_path = 'request_input', sheet_name = None):

        if file_path == 'request_input':
            file_path = input('File path: ')
        
        path_obj = Path(file_path)
        suffix = path_obj.suffix

        if path_obj.exists() == True:

            if suffix.strip('.') == 'xlsx':
                return self.import_excel(file_path, sheet_name)
            
            if suffix.strip('.') == 'csv':
                return self.import_csv(file_path)
            
            if suffix.strip('.') == 'json':
                return self.import_json(file_path)
        
        else:
            raise ValueError('File does not exist')
    
    def from_file(file_path = 'request_input', sheet_name = None): # type: ignore

        if file_path == 'request_input':
            file_path = input('File path: ')
        
        results_table = Results()

        path_obj = Path(file_path)  # type: ignore
        suffix = path_obj.suffix

        if path_obj.exists() == True:

            if suffix.strip('.') == 'xlsx':
                return results_table.import_excel(file_path, sheet_name)  # type: ignore
            
            if suffix.strip('.') == 'csv':
                return results_table.import_csv(file_path)  # type: ignore
            
            if suffix.strip('.') == 'json':
                return results_table.import_json(file_path) # type: ignore
        
        else:
            raise ValueError('File does not exist')

    def import_jstor(self, file_path = 'request_input', drop_empty_rows = False, drop_duplicates = False, update_work_ids = True):

        df = import_jstor(file_path = file_path)
        self.add_dataframe(dataframe=df, drop_empty_rows = drop_empty_rows, drop_duplicates = drop_duplicates, update_work_ids = update_work_ids)

    def search_field(self, field = 'request_input', any_kwds = 'request_input', all_kwds = None, not_kwds = None, case_sensitive = False, output = 'Results'):
        
        if field == 'request_input':
            field = input('Field: ')
        
        if any_kwds == 'request_input':
            any_kwds = input('Any keywords: ')
            any_kwds = any_kwds.strip().split(',')
            any_kwds = [i.strip() for i in any_kwds]

        contains_df = pd.DataFrame()
        not_kwds_df= pd.DataFrame()
        
        df = self.copy(deep=True)
        
        df[field] =  df[field].astype('str')

        if case_sensitive == False:
            df[field] = df[field].str.lower()
        
        if any_kwds != None:
                
            if (type(any_kwds) != list) and (type(any_kwds) != str):
                raise TypeError('"any_kwds" must be a string or list')

            if type(any_kwds) == str:
                any_kwds = [any_kwds]

            if case_sensitive == False:
                any_kwds = pd.Series(any_kwds).str.lower().to_list()

            for item in any_kwds:

                rows = df[df[field].str.contains(item) == True]
                contains_df = pd.concat([contains_df, rows])
                
        if not_kwds != None:

                if type(not_kwds) == str:
                    not_kwds = [not_kwds]

                if type(not_kwds) != list:
                    raise TypeError('"not_kwds" must be a string or list')

                if case_sensitive == False:
                    not_kwds = pd.Series(not_kwds).str.lower().to_list()

                for item in not_kwds:
                    
                        rows = df[df[field].str.contains(item) == False]
                        not_kwds = pd.concat([not_kwds, rows]) # type: ignore
        
        
        combined_df = pd.concat([contains_df, not_kwds_df])
        
        if all_kwds != None:
                
            if (type(all_kwds) != list) and (type(all_kwds) != str):
                raise TypeError('"all_kwds" must be a string or list')

            if type(all_kwds) == str:
                any_kwds = any_kwds.strip().split(',') # type: ignore
                any_kwds = [i.strip() for i in any_kwds]

            if case_sensitive == False:
                all_kwds = pd.Series(all_kwds).str.lower().to_list()
            
            for item in all_kwds:
                    
                combined_df = combined_df[combined_df[field].str.contains(item)]
        
        if output.lower() == 'results':

            masked = self.loc[combined_df.index]
            return masked

        else:
            if (output.lower() == 'dataframe') or (output.lower() == 'pandas.dataframe') or (output.lower() == 'pd.dataframe'):
                return combined_df

    def search(self, fields = 'all', any_kwds = 'request_input', all_kwds = None, not_kwds = None, case_sensitive = False, output = 'Results'):

        if any_kwds == 'request_input':
            any_kwds = input('Any keywords: ')
            any_kwds = any_kwds.strip().split(',')
            any_kwds = [i.strip() for i in any_kwds]
    
        if all_kwds != None:
                
            if (type(all_kwds) != list) and (type(all_kwds) != str):
                raise TypeError('"all_kwds" must be a string or list')

            if type(all_kwds) == str:
                any_kwds = any_kwds.strip().split(',') # type: ignore
                any_kwds = [i.strip() for i in any_kwds]

            if case_sensitive == False:
                all_kwds = pd.Series(all_kwds).str.lower().to_list()

        global results_cols

        output_df = pd.DataFrame(columns = results_cols, dtype=object)

        if fields == 'all':
            
            output_df = pd.DataFrame(columns = self.columns.to_list(), dtype=object)
            
            for col in self.columns:
                rows = self.search_field(field = col, any_kwds = any_kwds, not_kwds = not_kwds, case_sensitive = case_sensitive, output = 'pandas.dataframe') # type: ignore
                output_df = pd.concat([output_df, rows])
            
            output_df = output_df[~output_df.index.duplicated(keep = 'first') == True]
            
            if all_kwds != None:
                
                for item in all_kwds:
                    for i in output_df.index:
                        if True not in (output_df.loc[i].str.contains(item) == True).to_list():
                            output_df = output_df.drop(i)
    
        else:
            
            if (type(fields) == str) or ((type(fields) == list) and (len(fields) == 1)):

                output_df = self.search_field(field = fields, any_kwds = any_kwds, not_kwds = not_kwds, case_sensitive = case_sensitive) # type: ignore
                output_df = output_df[~output_df.index.duplicated(keep = 'first') == True] # type: ignore
                
                if all_kwds != None:
                    
                    for item in all_kwds:
                        
                        for i in output_df.index:
                            if True not in (output_df.loc[i].str.contains(item) == True).to_list():
                                output_df = output_df.drop(i)
        
        if output.lower() == 'results':
            masked = self.loc[output_df.index]
            return masked

        else:
            if (output.lower() == 'dataframe') or (output.lower() == 'pandas.dataframe') or (output.lower() == 'pd.dataframe'):
                return output_df
    
    def get_keywords(self):
    
        output = []

        for i in self['keywords']:
            if type(i) == str:
                i = strip_list_str(i)  # type: ignore
            
            if type(i) == list:
                output = output + i

        output = pd.Series(output).str.strip().str.lower()
        output = output.drop(output[output.values == 'none'].index).reset_index().drop('index', axis=1)[0] # type: ignore
        
        return output
    
    def get_keywords_list(self):        
        return self.get_keywords().to_list()
    
    def get_keywords_set(self):
        return set(self.get_keywords_list())
    
    def keyword_frequencies(self):
        return self.get_keywords().value_counts()

    def keyword_stats(self):
        return self.keyword_frequencies().describe()
    
    def get_titles_words(self, ignore_stopwords = True):

        df = self.copy(deep=True)
        df['title'] = df['title'].astype(str).str.lower().str.strip()

        output = []

        for i in df['title']:
            title = i.strip().replace("'", "").replace("`", "").replace(':', ' ').replace('-', ' ')
            words = list(word_tokenize(title))
            output = output + words
        
    
        if ignore_stopwords == True:
            
            global stopwords
            output = [i for i in output if i not in stopwords['all']]
        
        return output
    
    def get_titles_words_set(self, ignore_stopwords = True):
        return set(self.get_titles_words(ignore_stopwords = ignore_stopwords))
    
    def title_word_frequencies(self, ignore_stopwords = True):
        return pd.Series(self.get_titles_words(ignore_stopwords = ignore_stopwords)).value_counts()
                                               
    def title_words_stats(self, ignore_stopwords = True):
        return self.title_word_frequencies(ignore_stopwords = ignore_stopwords).describe()
    
    def drop_containing_keywords(self, keywords):
        
        if type(keywords) == str:
            keywords = [keywords]

        results = self.search(any_kwds = keywords).index # type: ignore
        return self.drop(index = results, axis=0).reset_index().drop('index', axis=1)

    def filter_by_keyword_frequency(self, cutoff = 3):

        keywords_freq = self.keyword_frequencies()

        frequent_kws = keywords_freq[keywords_freq.values > cutoff] # type: ignore
        frequent_kws = list(frequent_kws.index)
        
        output = pd.DataFrame(dtype=object)
        for i in frequent_kws:
            df = self.search(any_kwds = i).copy(deep=True) # type: ignore
            output = pd.concat([output, df])

        output = output.drop_duplicates('title')
        output = output.reset_index().drop('index', axis=1)

        print(f'Keywords: {frequent_kws}')

        return self.loc[output.index]
    
    def has_citations_data(self):
        return self[~self['citations_data'].isna()]

    def has(self, column):
        return self[~self[column].isna()]
    
    def contains(self, query: str = 'request_input', ignore_case: bool = True) -> bool:

        if query == 'request_input':
            query = input('Search query').strip()

        query = str(query).strip()

        all_str = self.copy(deep=True).astype(str)
        
        if ignore_case == True:
            query = query.lower()
            

        cols = all_str.columns

        for c in cols:

            if ignore_case == True:
                all_str[c] = all_str[c].str.lower()

            if c == 'work_id':
                res = all_str[all_str[c].str.contains(query)]
                if len(res) > 0:
                    return True
            
            if c == 'title':
                res = all_str[all_str[c] == query]
                if len(res) > 0:
                    return True
            
            if c == 'date':
                res = all_str[all_str[c].str.contains(query)]
                if len(res) > 0:
                    return True
            
            if c == 'source':
                res = all_str[all_str[c].str.contains(query)]
                if len(res) > 0:
                    return True
            
            if c == 'publisher':
                res = all_str[all_str[c].str.contains(query)]
                if len(res) > 0:
                    return True
            
            if c == 'funder':
                res = all_str[all_str[c].str.contains(query)]
                if len(res) > 0:
                    return True
            
            if c == 'keywords':
                res = all_str[all_str[c].str.contains(query)]
                if len(res) > 0:
                    return True
            
            if c == 'doi':
                res = all_str[all_str[c].str.contains(query)]
                if len(res) > 0:
                    return True
            
            if c == 'isbn':
                res = all_str[all_str[c] == query]
                if len(res) > 0:
                    return True
            
            if c == 'issn':
                res = all_str[all_str[c] == query]
                if len(res) > 0:
                    return True
            
            if c == 'link':
                res = all_str[all_str[c] == query]
                if len(res) > 0:
                    return True

        return False

    def mask_affiliations(self, query: str = 'request_input', ignore_case: bool = True):

        if query == 'request_input':
            query = input('Search query: ').strip()

        query = str(query).strip()

        def affil_masker(authors):
            if 'mask_entities' in authors.__dir__():
                masked = authors.mask_entities(column='affiliations', query=query, ignore_case=ignore_case)
                res = len(masked) > 0
                
            else:
                res = False

            return res

        
        masked = self[self['authors'].apply(affil_masker)]

        return masked

    def mask_entities(self, column, query: str = 'request_input', ignore_case: bool = True):

        if query == 'request_input':
            query = input('Search query').strip()

        query = str(query).strip()

        def entity_masker(entities):
            
            if 'contains' in entities.__dir__():
                res = entities.contains(query=query, ignore_case=ignore_case)
                
            else:
                res = False

            return res
        
        masked = self[self[column].apply(entity_masker)]

        return masked

    def format_funders(self, use_api: bool = False):

        try:
            funders = self['funder'].apply(format_funders) # type: ignore
        except:
            funders = self['funder']

        self['funder'] = funders

Entity.publications = Results() # type: ignore
Funder.publications = Results() # type: ignore


